import bibliograph as bg
import pandas as pd


class AssertionsNotFoundError(AttributeError):
    pass


class NodesNotFoundError(AttributeError):
    pass


class TextNet():

    def __init__(
        self,
        assertions=None,
        strings=None,
        nodes=None,
        edges=None,
        node_types=None,
        link_types=None,
        assertion_tags=None,
        edge_tags=None,
        node_metadata_tables=None,
        big_id_dtype=pd.Int32Dtype(),
        small_id_dtype=pd.Int8Dtype()
    ):
        self.assertions = assertions
        self.strings = strings
        self.nodes = nodes
        self.edges = edges
        self.node_types = node_types
        self.link_types = link_types
        self.assertion_tags = assertion_tags
        self.edge_tags = edge_tags
        node_metadata_tables = node_metadata_tables
        self.big_id_dtype = big_id_dtype
        self.small_id_dtype = small_id_dtype

        self._string_side_tables = [
            'strings', 'assertions', 'link_types', 'assertion_tags'
        ]
        self._node_side_tables = [
            'nodes', 'edges', 'node_types', 'edge_tags'
        ]

        if node_metadata_tables is None:
            self.node_metadata_tables = {}

        self._assertions_dtypes = {
            'inp_string_id': self.big_id_dtype,
            'src_string_id': self.big_id_dtype,
            'tgt_string_id': self.big_id_dtype,
            'ref_string_id': self.big_id_dtype,
            'link_type_id': self.small_id_dtype,
            'date_inserted': 'object',
            'date_modified': 'object'
        }
        self._assertions_index_dtype = self.big_id_dtype

        self._strings_dtypes = {
            'node_id': self.big_id_dtype,
            'string': 'object',
            'date_inserted': 'object',
            'date_modified': 'object'
        }
        self._strings_index_dtype = self.big_id_dtype

        self._nodes_dtypes = {
            'node_type_id': self.small_id_dtype,
            'name_string_id': self.big_id_dtype,
            'abbr_string_id': self.big_id_dtype,
            'date_inserted': 'object',
            'date_modified': 'object'
        }
        self._nodes_index_dtype = self.big_id_dtype

        self._edges_dtypes = {
            'src_node_id': self.big_id_dtype,
            'tgt_node_id': self.big_id_dtype,
            'ref_node_id': self.big_id_dtype,
            'link_type_id': self.small_id_dtype,
            'date_inserted': 'object',
            'date_modified': 'object'
        }
        self._edges_index_dtype = self.big_id_dtype

        self._node_types_dtypes = {
            'node_type': 'object',
            'description': 'object',
            'null_type': bool,
            'has_metadata': bool
        }
        self._node_types_index_dtype = self.small_id_dtype

        self._link_types_dtypes = {
            'link_type': 'object',
            'description': 'object',
            'null_type': bool
        }
        self._link_types_index_dtype = self.small_id_dtype

        self._assertion_tags_dtypes = {
            'assertion_id': self.big_id_dtype,
            'tag_string_id': self.big_id_dtype
        }
        self._assertion_tags_index_dtype = self.big_id_dtype

        self._edge_tags_dtypes = {
            'edge_id': self.big_id_dtype,
            'tag_string_id': self.big_id_dtype
        }
        self._edge_tags_index_dtype = self.big_id_dtype

    def __getattr__(self, attr):

        try:

            return self.__getattribute__(attr)

        except AttributeError as error:

            try:

                return self.node_metadata_tables[attr]

            except KeyError:

                if attr in self._string_side_tables:
                    raise AssertionsNotFoundError(
                        'assertions and strings not initialized for '
                        'this TextNet'
                    )

                elif attr in self._node_side_tables:
                    raise NodesNotFoundError(
                        'nodes and edges not initialized for '
                        'this TextNet'
                    )

                else:
                    raise error

    def _insert_type(
        self,
        name,
        description,
        null_type,
        has_metadata,
        node_or_link,
        overwrite_description
    ):

        table_name = node_or_link + '_types'
        column_name = node_or_link + '_type'
        existing_table = self.__getattribute__(table_name)

        if (node_or_link == 'link') and (has_metadata is not None):
            raise ValueError(
                'Cannot set "has_metadata" flag for link types'
            )

        if name in existing_table[column_name].array:

            if pd.notna(description) and overwrite_description:
                existing_row = (existing_table[column_name] == name)
                existing_table.loc[existing_row, 'description'] = description
                self.__setattr__(table_name, existing_table)

            return self.id_lookup(table_name, name)

        else:

            new_row = {
                column_name: name,
                'description': description,
                'null_type': bool(null_type)
            }
            new_row = bg.util.normalize_types(new_row, existing_table)

            if node_or_link == 'node':
                new_row['has_metadata'] = bool(has_metadata)

            self.__setattr__(table_name, pd.concat([existing_table, new_row]))

            return new_row.index[0]

    def _get_null_type_ids(self, table_name):
        table = self.__getattribute__(table_name)
        return table.loc[table['null_type']].index

    def _reset_table_dtypes(self, table_name):
        table_dtypes = self.__getattr__('_{}_dtypes'.format(table_name))
        index_dtype = self.__getattr__('_{}_index_dtype'.format(table_name))

        table = self.__getattr__(table_name)

        table = table.astype(table_dtypes)
        table.index = table.index.astype(index_dtype)
        table = table[table_dtypes.keys()]
        table = table.fillna(pd.NA)

        self.__setattr__(table_name, table)

    def get_null_link_type_ids(self):
        return self._get_null_type_ids('link_types')

    def get_null_node_type_ids(self):
        return self._get_null_type_ids('node_types')

    def get_nodes_with_null_types(self):
        null_type_ids = self.get_null_node_type_ids()
        return self.nodes.loc[self.nodes['node_type_id'].isin(null_type_ids)]

    def get_links_with_null_types(self):
        null_type_ids = self.get_null_link_type_ids()
        return self.links.loc[self.links['link_type_id'].isin(null_type_ids)]

    def get_node_types_by_node_id(self, node_ids):
        return self.nodes.loc[node_ids, 'node_type_id'].map(
            self.node_types['node_type']
        )

    def get_node_types_by_string_id(self, string_ids):
        node_ids = self.strings.loc[string_ids, 'node_id']
        return self.nodes.loc[node_ids, 'node_type_id'].map(
            self.node_types['node_type']
        )

    def insert_link_type(
        self,
        name,
        description=pd.NA,
        null_type=False,
        overwrite_description=False
    ):
        return self._insert_type(
            name,
            description,
            null_type,
            None,
            'link',
            overwrite_description
        )

    def insert_node_type(
        self,
        name,
        description=pd.NA,
        null_type=False,
        has_metadata=False,
        overwrite_description=False
    ):
        return self._insert_type(
            name,
            description,
            null_type,
            has_metadata,
            'node',
            overwrite_description
        )

    def insert_metadata_table(self, node_type, metadata):

        node_type_id = self.id_lookup(
            'node_types',
            node_type,
            column_label='node_type'
        )

        if self.nodes is not None:

            node_id = self.nodes.query('node_type_id == @node_type_id').index

            if len(node_id) > 1:
                raise NotImplementedError(
                    "Can't handle metadata for multiple nodes of the "
                    "same type"
                )

            metadata_table = {
                'node_id': node_id[0],
                'node_type_id': node_type_id
            }
        
        elif self.strings is not None:

            string_id = self.strings.query('node_type_id == @node_type_id')
            string_id = string_id.index

            if len(string_id) > 1:
                raise NotImplementedError(
                    "Can't handle metadata for multiple nodes of the "
                    "same type"
                )

            metadata_table = {
                'string_id': string_id[0],
                'node_type_id': node_type_id
            }

        else:

            raise ValueError(
                'Cannot insert a metadata table. TextNet.nodes and '
                'TextNet.strings are both None.'
            )

        metadata_table.update(metadata)

        metadata_table = {
            k: ([v] if v is not None else pd.NA)
            for k, v in metadata_table.items()
        }

        self.node_metadata_tables[node_type] = pd.DataFrame(
            metadata_table,
            index=pd.Index([0], dtype=self.small_id_dtype)
        )

        self.node_types.loc[node_type_id, 'has_metadata'] = True

    def id_lookup(self, attr, string, column_label=None):
        '''
        Take the name of an attribute of ParsedShorthand. If the
        attribute is a pandas Series or DataFrame, return the numerical
        index of the input string within the attribute.

        Parameters
        ----------
        attr : str
            Name of an attribute of ParsedShorthand

        string : str
            String value to retrieve index for

        column_label : str, default 'string'
            Column label to index if ParsedShorthand.attr is a
            DataFrame. Default to 'string' because at the moment
            ParsedShorthand.strings is the only DataFrame available.
        '''

        attribute = self.__getattribute__(attr)

        default_columns = {
            'strings': 'string',
            'node_types': 'node_type',
            'link_types': 'link_type'
        }

        try:
            # If this assertion passes, assume attribute is a Series
            assert attribute.str

            element = attribute.loc[attribute == string]

        except AttributeError:
            # Otherwise assume attribute is a DataFrame
            if column_label is None:

                if attr not in default_columns.keys():
                    raise ValueError(
                        'Must use column_label keyword when indexing {}'
                        .format(attr)
                    )

                else:
                    column_label = default_columns[attr]

            element = attribute.loc[attribute[column_label] == string]

        length = len(element)

        if length == 1:
            return element.index[0]
        elif length == 0:
            raise KeyError(string)
        elif length > 1:
            raise ValueError('{} index not unique!'.format(attr))

    def map_string_id_to_node_type(self, string_id):

        try:
            int(string_id)

            output = self.node_types.loc[
                self.strings.loc[string_id, 'node_type_id'],
                'node_type'
            ]

            return output[0]

        except TypeError:

            try:
                output = string_id.map(self.strings['node_type_id'])
                return output.map(self.node_types['node_type'])

            except KeyError:
                output = string_id.map(self.strings['node_id'])
                output = output.map(self.nodes['node_type_id'])
                return output.map(self.node_types['node_type'])

    def reset_assertions_dtypes(self):
        self._reset_table_dtypes('assertions')

    def reset_strings_dtypes(self):
        self._reset_table_dtypes('strings')

    def reset_nodes_dtypes(self):
        self._reset_table_dtypes('nodes')

    def reset_edges_dtypes(self):
        self._reset_table_dtypes('edges')

    def reset_node_types_dtypes(self):
        self._reset_table_dtypes('node_types')

    def reset_link_types_dtypes(self):
        self._reset_table_dtypes('link_types')

    def reset_assertion_tags_dtypes(self):
        self._reset_table_dtypes('assertion_tags')

    def reset_edge_tags_dtypes(self):
        self._reset_table_dtypes('edge_tags')

    def resolve_assertions(self, node_types=True, tags=True):
        '''
        Get a copy of the assertions frame with all integer ID elements
        replaced by the string values they represent

        Parameters
        ----------
        node_types : bool, default True
            If True, add columns for the node types of the source,
            target, and reference strings (three additional columns).

        tags : bool, default True
            If True, add a column for assertion tags. All tags for a
            link are joined into a single string separated by spaces.

        Returns
        -------
        pandas.DataFrame
        '''

        assertions = self.assertions
        string_map = self.strings['string']
        lt_map = self.link_types['link_type']

        resolved = pd.DataFrame(
            {'inp_string': assertions['inp_string_id'].map(string_map),
             'src_string': assertions['src_string_id'].map(string_map),
             'tgt_string': assertions['tgt_string_id'].map(string_map),
             'ref_string': assertions['ref_string_id'].map(string_map),
             'link_type': assertions['link_type_id'].map(lt_map),
             'date_inserted': assertions['date_inserted'],
             'date_modified': assertions['date_modified']},
            index=assertions.index
        )

        if node_types:
            resolved['inp_node_type'] = self.map_string_id_to_node_type(
                assertions['inp_string_id']
            )
            resolved['src_node_type'] = self.map_string_id_to_node_type(
                assertions['src_string_id']
            )
            resolved['tgt_node_type'] = self.map_string_id_to_node_type(
                assertions['tgt_string_id']
            )
            resolved['ref_node_type'] = self.map_string_id_to_node_type(
                assertions['ref_string_id']
            )

        if tags is True:
            # Resolve link tags as space-delimited lists
            tags = self.assertion_tags.groupby('assertion_id')
            tags = tags.apply(
                lambda x:
                ' '.join(self.strings.loc[x['tag_string_id'], 'string'])
            )

            resolved = resolved.join(tags.rename('tags'))

        return resolved.fillna(pd.NA)

    def resolve_edges(
        self,
        string_type='name',
        node_types=True,
        tags=True
    ):
        '''
        Get a copy of the edges frame with all integer ID elements
        replaced by the string values they represent

        Parameters
        ----------
        node_types : bool, default True
            If True, add columns for the node types of the source,
            target, and reference strings (three additional columns).

        tags : bool, default True
            If True, add a column for assertion tags. All tags for a
            link are joined into a single string separated by spaces.

        Returns
        -------
        pandas.DataFrame
        '''

        edges = self.edges
        lt_map = self.link_types['link_type']

        def map_to_strings(node_ids):
            string_ids = node_ids.map(self.nodes[string_type + '_string_id'])
            return string_ids.map(self.strings['string'])

        resolved = pd.DataFrame(
            {'src_string': map_to_strings(edges['src_node_id']),
             'tgt_string': map_to_strings(edges['tgt_node_id']),
             'ref_string': map_to_strings(edges['ref_node_id']),
             'link_type': edges['link_type_id'].map(lt_map),
             'date_inserted': edges['date_inserted'],
             'date_modified': edges['date_modified']},
            index=edges.index
        )

        if node_types:
            resolved['src_node_type'] = self.get_node_types_by_node_id(
                edges['src_node_id']
            ).array
            resolved['tgt_node_type'] = self.get_node_types_by_node_id(
                edges['tgt_node_id']
            ).array
            resolved['ref_node_type'] = self.get_node_types_by_node_id(
                edges['ref_node_id']
            ).array

        if tags is True and not self.edge_tags.empty:
            # Resolve link tags as space-delimited lists
            tags = self.edge_tags.groupby('edge_id')
            tags = tags.apply(
                lambda x:
                ' '.join(self.strings.loc[x['tag_string_id'], 'string'])
            )

            resolved = pd.concat(
                [resolved, tags.rename('tags')],
                axis='columns'
            )

        return resolved.fillna(pd.NA)

    def resolve_nodes(self):
        '''
        Get a copy of the nodes frame with all integer ID elements
        replaced by the string values they represent

        Returns
        -------
        pandas.DataFrame
            Same shape and index as the strings frame.
        '''

        nodes = self.nodes
        node_types = nodes['node_type_id'].map(self.node_types['node_type'])
        name_strings = nodes['name_string_id'].map(self.strings['string'])
        abbr_strings = nodes['abbr_string_id'].map(self.strings['string'])

        resolved = pd.DataFrame(
            {'node_type': node_types,
             'name_string': name_strings,
             'abbr_string': abbr_strings,
             'date_inserted': nodes['date_inserted'],
             'date_modified': nodes['date_modified']},
            index=nodes.index
        )

        return resolved.fillna(pd.NA)

    def resolve_strings(self):
        '''
        Get a copy of the strings frame with all integer ID elements
        replaced by the string values they represent

        Returns
        -------
        pandas.DataFrame
            Same shape and index as the strings frame.
        '''

        resolved = self.strings.copy()

        try:

            resolved['node_type'] = resolved['node_id'].map(
                self.nodes['node_type_id']
            )
            resolved['node_type'] = resolved['node_type'].map(
                self.node_types['node_type']
            )
            return resolved[[
                'node_id',
                'string',
                'node_type',
                'date_inserted',
                'date_modified'
            ]]

        except AttributeError:

            resolved['node_type_id'] = self.strings['node_type_id'].map(
                self.node_types['node_type']
            )
            return resolved.rename(columns={'node_type_id': 'node_type'})

    def select_strings_by_node_type(self, node_type):

        # if node_type is list-like, get the set
        if bg.util.iterable_not_string(node_type):

            node_type_id = [
                self.id_lookup('node_types', nt) for nt in node_type
            ]

            if self.nodes is None:
                selection = self.strings.query(
                    'node_type_id.isin(@node_type_id)'
                )
            else:
                selection = self.nodes.query(
                    'node_type_id.isin(@node_type_id)'
                )

        # if not list-like, first assume node_type is a string
        try:

            assert node_type.casefold()

            node_type_id = self.id_lookup('node_types', node_type)

            if self.nodes is None:
                selection = self.strings.query(
                    'node_type_id == @node_type_id'
                )
            else:
                selection = self.nodes.query(
                    'node_type_id == @node_type_id'
                )

        # if node_type doesn't have casefold, assume it's a numeric
        # node type ID
        except AttributeError:

            if self.nodes is None:
                selection = self.strings.query(
                    'node_type_id == @node_type'
                )
            else:
                selection = self.nodes.query(
                    'node_type_id == @node_type'
                )

        if self.nodes is None:
            return selection

        else:
            string_index = selection.index
            return self.strings.query('node_id.isin(@string_index)')



import bibliograph as bg
import pandas as pd


def _collapse_columns(df, columns):

    if bg.util.iterable_not_string(columns):
        if len(columns) > 1:
            if not df[columns].count(axis=1).lt(2).all():
                raise ValueError(
                    'found multiple values in a single row for columns '
                    '{}'.format(columns)
                )
            return df[columns].ffill(axis=1).dropna(axis=1)

        elif len(columns) == 1:
            return pd.DataFrame(df[columns])

        else:
            return pd.DataFrame(dtype='object')

    elif (type(columns) == str):
        return df[columns]

    else:
        raise ValueError('unrecognized input value for columns')


def _concat_list_item_elements(list_elements):

    sep = bg.util.get_single_value(list_elements, 'list_delimiter')
    item_label = bg.util.get_single_value(list_elements, 'item_label')

    item_string = sep.join(
        list_elements.sort_values(by='list_position')['tgt_string']
    )

    item = pd.DataFrame(
        {
            'item_position': list_elements.name,
            'tgt_string': item_string,
            'item_label': item_label
        },
        index=[list_elements.index.min()]
    )

    return item


def _synthesize_entries_from_links(
    group,
    entry_syntax,
    prefix_label_map,
    strings,
):
    '''
    Check if a set of links with a common source string ID have links
    like a shorthand entry with the given syntax. If so, return a pandas
    DataFrame with string values for each item in the entry.

    Parameters
    ----------
    group : pandas.DataFrame
        The set of all links whose source is a given src_string_id.
        Has columns
        [
            'src_string_id',
            'tgt_string_id',
            'link_type_id',
            'list_position',
            'ref_string_id'
        ]
        every row should have the same value in column 'src_string_id'

    entry_syntax : pandas.DataFrame
        syntax definition for a single entry prefix. has columns
        [
            'item_node_type',
            'item_link_type',
            'list_delimiter',
            'item_position',
            'item_label'
        ]
        where the node and link types are integer IDs, item_position is
        numeric but string-valued, and item_label is string-valued and
        alphanumeric.

    prefix_label_map : pandas.Series
        A map between positional item labels (the index) and
        alphanumeric item prefixes (the array). The array should be
        sorted by the index and then by the position of the prefix
        within the list of prefixes in the entry syntax file.

    strings : pandas.DataFrame
        has columns ['string', 'node_type_id']

    Returns
    -------
    pandas.DataFrame
    '''

    # Cache an empty return value
    return_cols = [
        'tgt_string',
        'item_label',
        'item_position',
        'item_prefix_separator'
    ]
    empty = pd.DataFrame(columns=return_cols)

    # The minimum number of links required for this source string to
    # represent an instance of the entry prefix with the given syntax is
    # the number of unprefixed item positions in the entry syntax. They
    # are denoted with integer item labels. The entry_prefix.notna()
    # query excludes entry syntax rows that were added to account for
    # links to the null node.
    min_links = entry_syntax.query('entry_prefix.notna()')
    min_links = min_links.query('item_label.str.isdigit()')
    min_links = len(min_links)
    if len(group) < min_links:
        return empty

    # Reduce group dataframe to relevant columns only, and make a copy
    # so we don't mutate an argument.
    group = group[['tgt_string_id', 'link_type_id', 'list_position']].copy()

    # This source string might be an instance of an entry that should
    # have the requested prefix. To check, we need to get the node types
    # of the link targets and compare them to the node types and link
    # types listed in the entry syntax.
    group = group.merge(
        strings['node_type_id'].rename('tgt_node_type_id'),
        left_on='tgt_string_id',
        right_index=True,
        copy=False
    )

    # Hash the (node type, link type) pairs in the syntax and the link
    # group so we can treat the type pairs as single values.
    syntax_type_pairs = pd.util.hash_pandas_object(
        entry_syntax[['item_node_type', 'item_link_type']],
        index=False
    )
    group_type_pairs = pd.util.hash_pandas_object(
        group[['tgt_node_type_id', 'link_type_id']],
        index=False
    )

    # Replace the type columns in the input frames with the hashes
    entry_syntax['type_pair_hash'] = syntax_type_pairs.array
    group['type_pair_hash'] = group_type_pairs.array

    # The group must contain at least one type pair associated with each
    # item in the entry syntax to be an instance of an entry with the
    # requested prefix.
    type_pair_check = entry_syntax.groupby(by='item_position')
    # For each item position, check if any type pair from the syntax is
    # in the group.
    type_pair_check = type_pair_check.apply(
        lambda x: x['type_pair_hash'].isin(group['type_pair_hash']).any()
    )
    # If any of the items isn't represented in the group, return null.
    if not type_pair_check.all():
        return empty

    # We now assume this source string is one that would have the
    # requested prefix.

    # Merge the link group with the entry syntax so we have enough
    # information to get the link targets in order
    relevant_syntax_columns = [
        'type_pair_hash',
        'list_delimiter',
        'item_position',
        'item_label',
        'item_prefix_separator'
    ]
    group = group.merge(entry_syntax[relevant_syntax_columns])

    # Insert the string values of the link targets into the group
    group['tgt_string'] = group['tgt_string_id'].map(strings['string'])

    # For any items that are lists of elements, concatenate the elements
    # using the list delimiter in the entry syntax
    list_items = group.query('list_position.notna()')

    if not list_items.empty:
        list_items = list_items.groupby(by='item_position', group_keys=False)
        list_items = list_items.apply(_concat_list_item_elements)

        group = pd.concat([group.query('list_position.isna()'), list_items])

    # The group of links now has one row per item position except for
    # prefixed items. The group might contain links to multiple prefixed
    # items even though only one value per item can be represented in an
    # entry string. To print an entry, we select the first prefix in the
    # prefix list from the entry syntax that is present in the group of
    # links.

    # Insert the precedence of the prefixes in the group
    prefix_precedence = group['item_label'].map(
        pd.Series(
            range(len(prefix_label_map)),
            index=prefix_label_map.index,
            dtype=pd.Int32Dtype()
        )
    )
    group = pd.concat(
        [group, prefix_precedence.rename('prefix_precedence')],
        axis='columns'
    )

    # Select prefixed items out of the group
    selected_prefixed_items = group.query('~item_label.str.isdigit()')

    # If there are no prefixed items, return the target strings, their
    # item labels (which could be prefixes or numbers indicating entry
    # positions), their positions in the entry, and their prefix
    # separators
    if selected_prefixed_items.empty:
        return group[return_cols]

    # Group items by their position in the entry and select only the
    # item with the first prefix in the list of prefixes in the entry
    # syntax.
    selected_prefixed_items = selected_prefixed_items.groupby(
        by='item_position'
    )
    selected_prefixed_items = selected_prefixed_items.apply(
        lambda x:
        x.loc[x['prefix_precedence'].astype(pd.Int32Dtype()).idxmin()]
    )

    # Combine the unprefixed items with the selected prefixed items and
    # sort by their position in the entry
    group = pd.concat([
            group.query('item_label.str.isdigit()'),
            selected_prefixed_items
        ])
    group = group.sort_values(by='item_position')

    # Return the target strings, their item labels (which could be
    # prefixes or numbers indicating entry positions), their positions
    # in the entry, and their prefix separators
    return group[return_cols]


def _prefix_targets(row):
    '''
    Takes a row from a dataframe representing one item in an entry. The
    string representation of the item is the target of a link between
    the entry string and the item. If the item has a prefix, add the
    prefix to the target string along with the prefix separator as
    defined in the syntax. Return the prefixed or unmodified string
    along with its position in the entry.

    Parameters
    ----------
    row : pandas Series
        Index includes the labels [
            'tgt_string',
            'item_label',
            'item_position',
            'item_prefix_separator'
        ]

    Returns
    -------
    pandas Series
        Has index ['item_position', 'tgt_string']
    '''

    if not str.isdigit(row['item_label']):
        components = row[['item_label', 'item_prefix_separator', 'tgt_string']]
        return pd.Series(
            [row['item_position'], components.sum()],
            index=['item_position', 'tgt_string']
        )

    else:
        return row[['item_position', 'tgt_string']]


def _prefix_column(column, prefix_separators):
    '''
    Takes a column from a multiindexed dataframe representing an entry.
    The multiindex (and therefore the column name) is the pair
    (item label, item position). If the item label for this column is a
    prefix then add the prefix with a separator to each string in the
    column. Otherwise return the column unchanged.

    Parameters
    ----------
    column : pandas.Series
        Should be one column out of a pandas.DataFrame whose column
        labels are a multiindex.

    prefix_separators : pandas.Series
        Map between item labels (the index) and prefix separators (the
        array)

    Returns
    -------
    pandas.Series
        Has same indexes as column
    '''

    if not str.isdigit(column.name[0]):
        sep = prefix_separators.loc[column.name[0]]
        return column.name[0] + sep + column

    else:
        return column


class ParsedShorthand:
    '''
    ParsedShorthand has attributes containing the results of a
    parsing operation and methods to synthesize them in various ways.


    Parameters
    ----------
    All dataframe columns with labels ending "_id", as well as the
    column links['list_position'] are integer-valued. Series and columns
    with other labels are string-valued. Indexes are always
    integer-valued.

    strings : pandas.DataFrame
        has columns ['string', 'node_type_id']

    links : pandas.DataFrame
        has columns [
            'src_string_id',
            'tgt_string_id',
            'link_type_id',
            'list_position',
            'ref_string_id'
        ]

    link_tags : pandas.DataFrame
        has columns ['link_id', 'tag_string_id']

    node_types : pandas.Series

    link_types : pandas.Series

    entry_prefixes : pandas.Series

    item_labels : pandas.Series

    item_separator : str

    default_entry_prefix : str

    space_char : str

    na_string_values : str or list-like of str

    syntax_case_sensitive : bool
    '''

    def __init__(
            self,
            strings,
            links,
            link_tags,
            node_types,
            link_types,
            entry_prefixes,
            item_labels,
            item_separator,
            default_entry_prefix,
            space_char,
            comment_char,
            na_string_values,
            na_node_type,
            syntax_case_sensitive,
            allow_redundant_items
    ):

        [self.__setattr__(k, v) for k, v in locals().items() if k != 'self']

    def _get_entries_by_prefix(
        self,
        entry_syntax,
        source_string_id_subset
    ):

        # If this entry prefix isn't listed in the entry syntax, return
        # an empty series
        if entry_syntax.empty:
            return pd.Series(dtype='object')

        # Items in an entry might be linked to the null node rather than
        # the target defined in the syntax. We only care about this when
        # checking what kind of links are present for an entry in the
        # parsed data, so we add rows to the entry syntax with every
        # link type paired with a null item node type
        null_syntax = entry_syntax.copy()[[
            'item_label',
            'item_node_type',
            'item_link_type',
            'list_delimiter'
        ]]
        null_syntax['item_node_type'] = self.na_node_type
        null_syntax = null_syntax.drop_duplicates()
        null_syntax = null_syntax.dropna(
            subset=['item_node_type', 'item_link_type']
        )
        entry_syntax = pd.concat([entry_syntax, null_syntax])

        # Make a map from positional index labels to any associated item
        # prefixes
        prefix_label_map = entry_syntax.query('item_prefixes.notna()')
        prefix_label_map = pd.Series(
            prefix_label_map['item_prefixes'].array,
            index=prefix_label_map['item_label'].array
        )
        prefix_label_map = prefix_label_map.str.split().explode()

        # Mutate it into a map from item prefixes to positional index
        # labels
        prefix_label_map = pd.Series(
            prefix_label_map.index,
            index=prefix_label_map
        )

        # Make a map from positional index labels to any associated item
        # prefix separators
        sep_label_map = entry_syntax.query('item_prefix_separator.notna()')
        sep_label_map = pd.Series(
            sep_label_map['item_prefix_separator'].array,
            index=sep_label_map['item_label'].array
        )
        sep_label_map = prefix_label_map.map(sep_label_map)

        # Use the separator map to propagate prefix separators from the
        # syntax rows for the positional items to the syntax rows for
        # the prefixed items.
        prefix_separators = entry_syntax['item_label'].map(sep_label_map)
        entry_syntax.loc[:, 'item_prefix_separator'] = prefix_separators.array

        # Make a new column to represent item positions within the entry
        # separately from the item labels
        entry_syntax = pd.concat(
            [
                entry_syntax,
                entry_syntax['item_label'].rename('item_position')
            ],
            axis='columns'
        )

        # Insert item positions for items whose labels are alphanumeric
        # prefixes rather than positions
        entry_syntax['item_position'].update(
            entry_syntax['item_label'].map(prefix_label_map)
        )

        # Convert string-valued node types to integer IDs
        node_type_id_map = pd.Series(
            self.node_types.index,
            index=self.node_types
        )
        node_type_ids = entry_syntax['item_node_type'].map(node_type_id_map)
        entry_syntax.loc[:, 'item_node_type'] = node_type_ids.array

        # Convert string-valued link types to integer IDs
        link_type_id_map = pd.Series(
            self.link_types.index,
            index=self.link_types
        )
        link_type_ids = entry_syntax['item_link_type'].map(link_type_id_map)
        entry_syntax.loc[:, 'item_link_type'] = link_type_ids.array

        # Get the integer node type ID for this entry
        entry_node_type = bg.util.get_single_value(
            entry_syntax,
            'entry_node_type'
        )
        entry_node_type = self.id_lookup('node_types', entry_node_type)

        # Get string IDs for strings that have the node type ID for this
        # entry
        string_id_selection = self.strings.query(
            'node_type_id == @entry_node_type'
        )
        string_id_selection = string_id_selection.index

        # Select links whose sources are the selected string IDs
        source_selected = self.links['src_string_id'].isin(
            string_id_selection
        )

        # Restrict selection to set of allowed source strings if caller
        # provided one
        if source_string_id_subset is not None:
            source_string_id_subset = self.links['src_string_id'].isin(
                source_string_id_subset
            )
            source_selected = source_selected & source_string_id_subset

        # Restrict selection to the link types listed for items
        # in the syntax for this entry prefix
        link_type_selected = self.links['link_type_id'].isin(
            entry_syntax['item_link_type']
        )

        link_selected = source_selected & link_type_selected

        # If these criteria leave no links selected, return an empty
        # Series. Otherwise, get a subset of links that meet our
        # criteria.
        if not link_selected.any():
            return pd.Series(dtype='object')
        else:
            link_selection = self.links.loc[link_selected]

        # Group the links by source string. If a source has links like
        # the entry prefix in the syntax, generate strings representing
        # items for that entry.
        link_selection = link_selection.groupby(by='src_string_id')

        entry_strings = link_selection.apply(
            _synthesize_entries_from_links,
            entry_syntax,
            prefix_label_map,
            self.strings
        )

        return entry_strings

    def id_lookup(self, attr, string, column_label='string'):
        '''
        Take the name of an attribute of ParsedShorthand. If the
        attribute is a pandas Series or DataFrame, return the numerical
        index of the input string within the attribute.

        Parameters
        ----------
        attr : str
            Name of an attribute of ParsedShorthand

        string : str
            String value to retrieve index for

        column_label : str, default 'string'
            Column label to index if ParsedShorthand.attr is a
            DataFrame. Default to 'string' because at the moment
            ParsedShorthand.strings is the only DataFrame available.
        '''

        attribute = self.__getattribute__(attr)

        try:
            # If this assertion passes, assume attribute is a Series
            assert attribute.str

            element = attribute.loc[attribute == string]

        except AttributeError:
            # Otherwise assume attribute is a DataFrame
            if column_label == 'string' and attr != 'strings':
                raise ValueError(
                    'Must use column_label keyword when indexing a '
                    'DataFrame'
                )

            element = attribute.loc[attribute[column_label] == string]
        length = len(element)

        if length == 1:
            return element.index[0]
        elif length == 0:
            raise KeyError(string)
        elif length > 1:
            raise ValueError('{} index not unique!'.format(attr))

    def resolve_strings(self):
        '''
        Get a copy of the strings frame with all integer ID elements
        replaced by the string values they represent

        Returns
        -------
        pandas.DataFrame
            Same shape and index as the strings frame.
        '''

        resolved = self.strings.copy()

        resolved['node_type_id'] = self.strings['node_type_id'].map(
            self.node_types
        )
        resolved = resolved.rename(columns={'node_type_id': 'node_type'})

        return resolved

    def resolve_links(self, node_types=True, tags=True):
        '''
        Get a copy of the links frame with all integer ID elements
        replaced by the string values they represent

        Parameters
        ----------
        node_types : bool, default True
            If True, add columns for the node types of the source,
            target, and reference strings (three additional columns).

        tags : bool, default True
            If True, add a column for link tags. All tags for a link are
            joined into a single string separated by spaces.

        Returns
        -------
        pandas.DataFrame
        '''

        string_map = self.strings['string']

        resolved = pd.DataFrame(
            {'src_string': self.links['src_string_id'].map(string_map),
             'tgt_string': self.links['tgt_string_id'].map(string_map)},
            index=self.links.index
        )

        resolved['link_type'] = self.links['link_type_id'].map(
            self.link_types
        )

        resolved['ref_string'] = self.links['ref_string_id'].map(string_map)

        if node_types:
            resolved['src_node_type'] = self.links['src_string_id'].map(
                self.strings['node_type_id']
            )
            resolved['src_node_type'] = resolved['src_node_type'].map(
                self.node_types
            )

            resolved['tgt_node_type'] = self.links['tgt_string_id'].map(
                self.strings['node_type_id']
            )
            resolved['tgt_node_type'] = resolved['tgt_node_type'].map(
                self.node_types
            )

            resolved['ref_node_type'] = self.links['ref_string_id'].map(
                self.strings['node_type_id']
            )
            resolved['ref_node_type'] = resolved['ref_node_type'].map(
                self.node_types
            )

        if tags is True:
            # Resolve link tags as space-delimited lists
            tags = self.link_tags.groupby('link_id')
            tags = tags.apply(
                lambda x:
                ' '.join(self.strings.loc[x['tag_string_id'], 'string'])
            )

        return resolved

    def synthesize_shorthand_entries(
        self,
        entry_prefix=None,
        entry_node_type=None,
        sort_by=None,
        sort_prefixes=True,
        sort_case_sensitive=True,
        source_string_id_subset=None,
        fill_spaces=False,
        missing_node_warning=True
    ):
        '''
        Takes an entry prefix and generates a pandas Series of string
        representations of each entry of that type in the parsed data.

        Parameters
        ----------
        entry_prefix : str
            prefix in the entry syntax

        entry_node_type : str
            entry node type in the entry syntax

        sort_by : str or list of str
            item columns on which the output should be sorted

        sort_prefixes : bool, default True
            If True, add item prefixes prior to sorting columns. If
            False, sort values then add any prefixes afterwards.

        source_string_id_subset : pandas.Series, default None
            Must be boolean dtype. If present, only generate output for
            entries whose source strings are in
            ParsedShorthand.strings.loc[source_string_id_subset]

        Returns
        -------
        pandas Series
            string-valued representations of a set of entries
        '''

        # Get the text of the entry syntax and validate it
        entry_syntax_node_type_ID = self.id_lookup(
            'node_types',
            'shorthand_entry_syntax'
        )
        entry_syntax = self.strings.query(
            'node_type_id == @entry_syntax_node_type_ID'
        )
        entry_syntax = bg.util.get_single_value(entry_syntax, 'string')
        entry_syntax = bg.syntax_parsing.validate_entry_syntax(
            entry_syntax,
            case_sensitive=self.syntax_case_sensitive
        )

        if entry_prefix is not None:
            # Select a single entry prefix out of the entry syntax
            entry_syntax = entry_syntax.query('entry_prefix == @entry_prefix')

            # If there's no node type for this entry prefix then no
            # synthesis is required. Read available entry strings
            # directly and return them.
            if entry_syntax['entry_node_type'].empty:
                if missing_node_warning:
                    print(
                        'Warning: requested entry type has no node '
                        'type in the entry syntax.\nSelecting'
                        'strings that begin with the entry prefix.'
                        '\nUnprefixed entries are ignored.'
                    )

                entry_strings = self.strings.query('node_type_id.isna()')
                prefix = entry_prefix + self.item_separator
                entry_strings = entry_strings.loc[
                    entry_strings['string'].str.startswith(prefix)
                ]

                entry_strings = entry_strings['string']

                if sort_by is not None:
                    entry_strings = entry_strings.sort_values()

                if fill_spaces:
                    entry_strings = entry_strings.str.replace(
                        ' ',
                        self.space_char,
                        regex=False
                    )

                return pd.Series(entry_strings.array)

            else:
                entry_strings = self._get_entries_by_prefix(
                    entry_syntax,
                    source_string_id_subset
                )

        elif entry_node_type is not None:
            # Select a single node type out of the entry syntax
            entry_syntax = entry_syntax.query(
                'entry_node_type == @entry_node_type'
            )

            # Synthesize entries for each entry prefix in the selection
            prefixes = entry_syntax['entry_prefix'].drop_duplicates()
            syntaxes = [
                entry_syntax.query('entry_prefix == @entry_prefix')
                for entry_prefix in prefixes
            ]
            entry_strings = [
                self._get_entries_by_prefix(syntax, source_string_id_subset)
                for syntax in syntaxes
            ]

            # Concatenate entries for the different prefixes
            entry_strings = pd.concat(entry_strings)

        if sort_by is None:
            # If we aren't sorting the output, insert prefixes for
            # prefixed strings and pivot the DataFrame so that each
            # row represents a single entry
            entry_strings = entry_strings.apply(
                _prefix_targets,
                axis='columns'
            )
            entry_strings = entry_strings.reset_index()
            entry_strings = entry_strings.pivot(
                index='src_string_id',
                columns='item_position',
                values='tgt_string'
            )

        else:

            # make sort_by list-like if it isn't already
            if not bg.util.iterable_not_string(sort_by):
                sort_by = [sort_by]

            # If we are sorting the output, make a map from item
            # labels to item prefix separators
            prefix_separators = entry_strings.copy()[
                ['item_prefix_separator', 'item_label']
            ]
            prefix_separators = prefix_separators.drop_duplicates()
            prefix_separators = prefix_separators.set_index('item_label')
            prefix_separators = prefix_separators.squeeze()

            # Pivot the entry strings into a DataFrame so that each
            # row represents a single entry. The column labels are a
            # multiindex containing both the item label, which could
            # be a prefix, and the item position, which is always a
            # string of digits
            entry_strings = entry_strings.reset_index()
            entry_strings = entry_strings.pivot(
                index='src_string_id',
                columns=['item_label', 'item_position'],
                values='tgt_string'
            )

            if len(entry_strings) == 1:
                order = entry_strings.index

            elif sort_prefixes:
                # If we're sorting the prefixes first, insert
                # prefixes
                entry_strings = entry_strings.apply(
                    _prefix_column,
                    args=(prefix_separators,)
                )

                # Get column(s) for the item position(s) we're
                # sorting on.
                order = [
                    entry_strings.loc[:, (slice(None), str(label))]
                    for label in sort_by
                ]

                # If the item to sort on is prefixed, then there is
                # one column for each (item prefix, item position)
                # pair, so we have to collapse those into a single
                # column before sorting
                order = [
                    _collapse_columns(part, part.columns)
                    for part in order
                ]

                # Make each item position into a single column,
                # optionally casefolding if the sort is not
                # case sensitive
                if sort_case_sensitive:
                    order = [part.squeeze() for part in order]
                else:
                    order = [
                        part.squeeze().str.casefold() for part in order
                    ]

                # concat the item position(s) we're sorting on into
                # a single dataframe
                order = pd.concat(order, axis='columns')

            else:
                # If we're sorting on unprefixed string values, do
                # the same operations described above, but get the
                # strings and sort them before adding the prefixes
                order = [
                    entry_strings.loc[:, (slice(None), str(label))]
                    for label in sort_by
                ]
                order = [
                    _collapse_columns(part, part.columns)
                    for part in order
                ]

                if sort_case_sensitive:
                    order = [part.squeeze() for part in order]
                else:
                    order = [
                        part.squeeze().str.casefold() for part in order
                    ]

                order = pd.concat(order, axis='columns')

                entry_strings = entry_strings.apply(
                    _prefix_column,
                    args=(prefix_separators,)
                )

            # Get an index of entry strings in the requested order
            order = order.sort_values(list(order.columns)).index

            # Group all columns by item position and collapse groups
            # into single columns.
            entry_strings = entry_strings.groupby(
                axis=1,
                level='item_position',
                group_keys=False
            )
            entry_strings = entry_strings.apply(
                lambda x: _collapse_columns(x, x.columns)
            )

            # Put the entries in the sorted order we generated above

            entry_strings = entry_strings.loc[order]

        entry_strings = entry_strings.apply(
            lambda x:
            x.str.replace(
                self.comment_char,
                '\\' + self.comment_char,
                regex=False
            )
        )

        entry_strings = entry_strings.apply(
            lambda x:
            x.str.replace(
                self.item_separator,
                '\\' + self.item_separator,
                regex=False
            )
        )

        # Join items by the item separator
        entry_strings = entry_strings.apply(
            lambda x: self.item_separator.join(x),
            axis='columns'
        )

        if fill_spaces:
            entry_strings = entry_strings.str.replace(
                ' ',
                self.space_char,
                regex=False
            )

        return pd.Series(entry_strings.array)



import bibliograph as bg
import pandas as pd
from io import StringIO
from pathlib import Path


def _extend_id_map(
    domain,
    existing_domain,
    existing_range=None,
    drop_na=True,
    **kwargs
):
    '''
    Map distinct values in a domain which are not in an existing domain
    to integers that do not overlap with an existing range. Additional
    keyword arguments are passed to the pandas.Series constructor when
    the map series is created.

    Parameters
    ----------
    domain : list-like (coercible to pandas.Series)
        Arbitrary set of values to map. May contain duplicates.

    existing_domain : list-like
        Set of values already present in a map from values to integers.

    existing_range : list-like or None, default None
        Range of integers already present in the range of a map. If
        None, assume existing_domain.index contains the existing range.

    drop_na : bool, default True
        Ignore null values and map only non-null values to integers.

    Examples
    --------
    >>> existing_domain = pd.Series(['a', 'a', 'b', 'f', 'b'])
    >>> new_domain = ['a', 'b', 'z', pd.NA]
    >>> _extend_id_map(new_domain,
                       existing_domain,
                       dtype=pd.UInt16Dtype())

    z    5
    dtype: UInt16

    >>> _extend_id_map(new_domain,
                       existing_domain,
                       dtype=pd.UInt16Dtype(),
                       drop_na=False)

    z       5
    <NA>    6
    dtype: UInt16
    '''
    # check if domain object has autocorr and loc attributes like a
    # pandas.Series and convert if not
    try:
        assert domain.autocorr
        assert domain.loc
        # make a copy so we can mutate one (potentially large) object
        # instead of creating additional references
        domain = domain.copy()
    except AttributeError:
        domain = pd.Series(domain)

    if drop_na:
        domain = domain.loc[~domain.isna()]

    domain_is_new = ~domain.isin(existing_domain)

    if domain_is_new.any():
        domain = domain.loc[domain_is_new].drop_duplicates()

        if existing_range is None:
            new_ids = bg.util.non_intersecting_sequence(
                len(domain),
                existing_domain.index
            )
        else:
            new_ids = bg.util.non_intersecting_sequence(
                len(domain),
                existing_range
            )

    else:
        domain = []
        new_ids = []

    return pd.Series(new_ids, index=domain, **kwargs)


class EntrySyntaxError(ValueError):
    pass


def _validate_entry_syntax_prefix_group(group, allow_redundant_items):

    bg.util.get_single_value(
        group,
        'entry_node_type',
        none_ok=True,
        group_key=group.name
    )

    no_entry_node_types = group['entry_node_type'].isna().all()
    any_entry_node_types = group['entry_node_type'].notna().any()
    no_link_types = group['item_link_type'].isna().all()
    any_link_types = group['item_link_type'].notna().any()

    msg = (
        'Error parsing syntax for entry_prefix {}. There are no values '
        'in column "{}" but column "{}" contains values.'
    )

    if no_entry_node_types and any_link_types:
        raise EntrySyntaxError(
            msg.format(group.name, 'entry_node_type', 'item_link_type')
        )
    if no_link_types and any_entry_node_types:
        raise EntrySyntaxError(
            msg.format(group.name, 'item_link_type', 'entry_node_type')
        )

    item_node_type_na = group['item_node_type'].isna()
    if item_node_type_na.any():
        if group.loc[item_node_type_na, 'item_link_type'].notna().any():
            raise EntrySyntaxError(
                'Error parsing syntax for entry_prefix {}. If any row '
                'has no value in column "item_node_type" then column '
                '"item_link_type" cannot contain a value in that '
                'row.'.format(group.name)
            )

    msg = (
        'Error parsing syntax for entry_prefix {}. If any row has a '
        'value in column "{}" then column "{}" must also contain a '
        'value in that row.'
    )

    list_delimiter_not_na = group['list_delimiter'].notna()
    if list_delimiter_not_na.any():
        if group.loc[list_delimiter_not_na, 'item_node_type'].isna().any():
            raise EntrySyntaxError(
                msg.format(group.name, 'list_delimiter', 'item_node_type')
            )

    item_prefix_separator_not_na = group['item_prefix_separator'].notna()
    if item_prefix_separator_not_na.any():
        has_item_prefix = group.loc[item_prefix_separator_not_na]
        if has_item_prefix.loc[:, 'item_prefixes'].isna().any():
            raise EntrySyntaxError(
                msg.format(
                    group.name, 'item_prefix_delimiter', 'item_prefixes'
                )
            )

        msg = (
            'Error parsing syntax for entry_prefix {}. If any row has '
            'a value in column "{}" then column "{}" cannot contain a '
            'value in that row.'
        )

        if has_item_prefix.loc[:, 'item_node_type'].notna().any():
            raise EntrySyntaxError(
                msg.format(
                    group.name, 'item_prefix_delimiter', 'item_node_type'
                )
            )
        if has_item_prefix.loc[:, 'item_link_type'].notna().any():
            raise EntrySyntaxError(
                msg.format(
                    group.name, 'item_prefix_delimiter', 'item_link_type'
                )
            )

        if group['item_label'].duplicated().any():
            raise EntrySyntaxError(
                'Error parsing labels for entry_prefix {}. Rows with '
                'the same entry prefix must have different item labels.'
                .format(group.name)
            )

    unprefixed_link_types = group.loc[
        ~item_prefix_separator_not_na,
        'item_link_type'
    ]
    if any_link_types and unprefixed_link_types.isna().any():
        raise EntrySyntaxError(
            'Error parsing labels for entry_prefix {}. If any rows '
            'have an item link type then all rows with no item prefix '
            'separator must also have an item link '
            'type.'.format(group.name)
        )

    item_data = ['item_node_type', 'item_link_type']
    duplicate_items = group[item_data].dropna(how='all').duplicated().any()
    if duplicate_items and not allow_redundant_items:
        raise EntrySyntaxError(
            'Error parsing syntax for entry prefix {}. Each row '
            'within a prefix group must have a unique pair of '
            'values for columns item_node_type and item_link_type '
            '(unless both values are null).'.format(group.name)
        )


def validate_entry_syntax(
    entry_syntax,
    case_sensitive,
    allow_redundant_items=False
):

    try:
        assert entry_syntax.casefold
        # if there's a casefold method then the entry syntax is a string

        if Path(entry_syntax).exists():
            # assume it's a csv file and read it
            entry_syntax = pd.read_csv(entry_syntax)
        else:
            # assume it's the string-valued contents of a csv
            # file and convert it to a pandas DataFrame
            with StringIO(entry_syntax) as stream:
                entry_syntax = pd.read_csv(stream)

    except AttributeError:
        raise ValueError(
            'entry_syntax must be a string representing a file path or the '
            'contents of a csv file. Got {}.'.format(type(entry_syntax))
        )

    # pandas converts numbers to numeric types with read_csv, but
    # we're treating everything as strings, so convert values
    entry_syntax = entry_syntax.applymap(str, na_action='ignore')

    # normalize column labels with unicode casefold (like making text
    # uniformly lower or upper case, but applies for the whole unicode
    # standard)
    entry_syntax.columns = entry_syntax.columns.str.casefold()

    if not case_sensitive:
        # Normalize all values with unicode casefold
        entry_syntax = entry_syntax.apply(
            lambda x: x.str.casefold() if x.notna().any() else x
        )

    if entry_syntax['item_label'].isna().any():
        raise ValueError(
            'Error parsing entry syntax. All rows must have a value in '
            'column "item_label"'
        )

    optional_cols = [
        'list_delimiter', 'item_prefix_separator', 'item_prefixes'
    ]
    add_cols = [
        col for col in optional_cols if col not in entry_syntax.columns
    ]
    entry_syntax = pd.concat(
        [
            entry_syntax,
            pd.DataFrame(columns=add_cols, index=entry_syntax.index)
        ],
        axis='columns')

    has_prefix = ('entry_prefix' in entry_syntax.columns)
    if has_prefix and entry_syntax['entry_prefix'].notna().any():
        entry_syntax.groupby('entry_prefix').apply(
            _validate_entry_syntax_prefix_group,
            allow_redundant_items
        )
    else:
        dummy = entry_syntax.copy()
        dummy.name = 'None'
        _validate_entry_syntax_prefix_group(dummy, allow_redundant_items)

    return entry_syntax


class LinkSyntaxError(ValueError):
    pass


def _validate_link_syntax(link_syntax, entry_syntax, case_sensitive):

    try:
        # assume link_syntax is a string if it has a casefold method
        assert link_syntax.casefold

        if Path(link_syntax).exists():
            # assume it's a csv file and read it
            link_syntax = pd.read_csv(link_syntax)
        else:
            # assume it's the string-valued contents of a csv
            # file and convert it to a pandas DataFrame
            with StringIO(link_syntax) as stream:
                link_syntax = pd.read_csv(stream)

    except AttributeError:
        raise ValueError(
            'link_syntax must be a string representing a file path or the '
            'contents of a csv file. Got {}.'.format(type(link_syntax))
        )

    # pandas converts numbers to numeric types with read_csv, but
    # we're treating everything as strings, so convert all values to str
    link_syntax = link_syntax.applymap(str, na_action='ignore')

    # normalize column labels with unicode casefold (like making text
    # uniformly lower or upper case, but applies for the whole unicode
    # standard)
    link_syntax.columns = link_syntax.columns.str.casefold()

    position_cols = [c for c in link_syntax.columns if 'position' in c]

    if case_sensitive:
        # normalize position codes and list modes with unicode casefold
        link_syntax[position_cols] = link_syntax[position_cols].apply(
            lambda x: x.str.casefold()
        )
        # normalize list modes with unicode casefold
        list_modes = link_syntax['list_mode'].str.casefold()
        link_syntax['list_mode'] = list_modes

    else:
        # normalize all values with unicode casefold
        link_syntax = link_syntax.apply(lambda x: x.str.casefold())

    required_columns = [
        'left_entry_prefix',
        'right_entry_prefix',
        'source_position',
        'target_position',
        'link_type',
        'list_mode',
        'dflt_ref_position'
    ]

    missing_columns = [
        c for c in required_columns if c not in link_syntax.columns
    ]

    if missing_columns != []:
        raise LinkSyntaxError(
            'Link syntax missing required columns {}'
            .format(missing_columns)
        )

    position_codes = link_syntax[position_cols].stack().dropna()
    # Casefolded position codes that do not begin with 'l' or 'r' are
    # invalid. Extract any invalid codes with a regular expression.
    invalid_pos_codes = position_codes.str.extract('^(?![lr])(.+)')
    invalid_pos_codes = invalid_pos_codes.dropna()
    if not invalid_pos_codes.empty:
        raise LinkSyntaxError(
            'Position codes in the link syntax must start with "R" '
            'or "L" (ignoring case). Found the following invalid '
            'position codes in the link syntax:\n>>> {}'
            .format(invalid_pos_codes)
        )

    # If there are any item labels in the link syntax that do not
    # exist in the entry syntax, throw an error.
    valid_item_labels = entry_syntax['item_label']
    item_labels = position_codes.str.slice(1)
    item_labels = item_labels.mask(item_labels == '')
    invalid_pos_codes = position_codes.loc[
        ~item_labels.isin(valid_item_labels) & ~item_labels.isna()
    ]
    if not invalid_pos_codes.empty:
        raise LinkSyntaxError(
            'Position codes in the link syntax must start with "R" '
            'or "L" (ignoring case) and continue with an item label '
            'defined in the entry syntax. Found the following position '
            'codes in the link syntax whose item labels do not '
            'exist in the entry syntax:\n>>> {}'.format(invalid_pos_codes)
        )

    # if list modes aren't one of [1:1, m:1, 1:m, m:m] throw an error
    valid_list_modes = ['1:1', 'm:1', '1:m', 'm:m']
    list_mode_notna = ~link_syntax['list_mode'].isna()
    invalid_list_modes = link_syntax['list_mode'].loc[
        list_mode_notna & ~link_syntax['list_mode'].isin(valid_list_modes)
    ]
    if not invalid_list_modes.empty:
        raise LinkSyntaxError(
            'List modes in the link syntax must be one of\n'
            '>>> {}\nFound the following invalid list modes in the '
            'link syntax:\n>>> {}'
            .format(valid_list_modes, list(invalid_list_modes))
        )

    return link_syntax


def parse_link_syntax(
    filepath_or_text,
    entry_syntax,
    entry_prefix_id_map,
    link_types,
    item_label_id_map,
    case_sensitive=True
):

    link_syntax = _validate_link_syntax(
        filepath_or_text,
        validate_entry_syntax(entry_syntax, case_sensitive),
        case_sensitive
    )

    # Select links with entry prefix pairs that exist in this data set
    prefix_cols = ['left_entry_prefix', 'right_entry_prefix']

    prefix_is_present = link_syntax[prefix_cols].isin(
        entry_prefix_id_map.index
    )

    prefix_is_present = prefix_is_present.all(axis=1)
    link_syntax = link_syntax.loc[prefix_is_present]

    # map entry prefixes from strings to integers and rename columns
    link_syntax[prefix_cols] = link_syntax[prefix_cols].apply(
        lambda x: x.map(entry_prefix_id_map)
    )
    link_syntax = link_syntax.rename(
        columns={
            'left_entry_prefix': 'left_entry_prefix_id',
            'right_entry_prefix': 'right_entry_prefix_id'
        }
    )

    # Get any new link types in the link syntax
    new_link_types = _extend_id_map(
        link_syntax['link_type'],
        link_types,
        dtype=link_types.index.dtype
    )
    new_link_types = pd.Series(new_link_types.index, index=new_link_types)

    # Add new link types
    link_types = pd.concat([link_types, new_link_types])

    # Map the link types from strings to integers
    link_syntax['link_type_id'] = link_syntax['link_type'].map(
        pd.Series(link_types.index, index=link_types.array)
    )
    link_syntax = link_syntax.drop('link_type', axis=1)

    # Links in the synthesized shorthand data are a relation between
    # four entities represented by integer IDs:
    #
    # src_string_id (string representing the source end of the link)
    # tgt_string_id (string representing the target end of the link)
    # ref_string_id (string representing the context of the link)
    # link_type_id (the link type)
    #
    # To generate these we first create a map between position codes in
    # the link syntax and prefixes for components of the link
    columns_and_component_prefixes = {
        'source_position': 'src_',
        'target_position': 'tgt_',
        'dflt_ref_position': 'ref_'
    }
    for column, component_prefix in columns_and_component_prefixes.items():
        position_code = link_syntax[column]

        # The first character in a position code is 'L' or 'R',
        # indicating the link component is either the left or the
        # right entry in the csv file. Slice these off the position
        # codes and create a column for the entry position for this
        # component of the link (source, target, or reference).
        entry_position = position_code.str.slice(0, 1).array
        link_syntax[component_prefix + 'csv_col'] = entry_position

        # The rest of the characters in a position code indicate which
        # item in the entry represents the link component (source,
        # target, or reference). These are either alphanumeric item
        # prefixes as defined in the entry syntax or zero-based
        # positional indices locating the item within the entry. Extract
        # the item labels with a regular expression which is null for
        # single-character position codes.
        item_label = position_code.str.extract('^.(.+)')
        # Create a column for the item label for this component of the
        # link component and drop the column for the position code.
        link_syntax[component_prefix + 'item_label'] = item_label
        link_syntax = link_syntax.drop(column, axis=1)

    # Map item labels to integers
    item_label_cols = [
        col for col in link_syntax.columns if 'item_label' in col
    ]
    link_syntax[item_label_cols] = link_syntax[item_label_cols].apply(
        lambda x: x.map(item_label_id_map)
    )

    return link_types, link_syntax



from bibliograph.ParsedShorthand import ParsedShorthand
from bibliograph.Shorthand import Shorthand
from bibliograph.TextNet import TextNet
from bibliograph.core import slurp_bibtex
from bibliograph.core import slurp_shorthand
from bibliograph.core import textnet_from_parsed_shorthand
import bibliograph.alias_generators as alias_generators
import bibliograph.entry_parsing as entry_parsing
import bibliograph.syntax_parsing as syntax_parsing
import bibliograph.util as util



import bibliograph as bg
import pandas as pd
from bibtexparser.bparser import BibTexParser
from io import StringIO


def test_manual_annotation_nodes_column_nan_states():

    tn = bg.slurp_shorthand(
        'bibliograph/test_data/manual_annotation.shnd',
        "bibliograph/resources/default_entry_syntax.csv",
        link_syntax_fname="bibliograph/resources/default_link_syntax.csv",
        syntax_case_sensitive=False,
        item_separator='__',
        default_entry_prefix='wrk',
        space_char='|',
        na_string_values=['!', 'x'],
        na_node_type='missing',
        skiprows=2,
        comment_char='#'
    )

    notna_subset = [
        'node_type_id', 'name_string_id', 'abbr_string_id', 'date_inserted'
    ]

    assert not tn.nodes[notna_subset].isna().any(axis=None)
    assert tn.nodes['date_modified'].isna().all()


def test_manual_annotation_edges_column_nan_states():

    tn = bg.slurp_shorthand(
        'bibliograph/test_data/manual_annotation.shnd',
        "bibliograph/resources/default_entry_syntax.csv",
        link_syntax_fname="bibliograph/resources/default_link_syntax.csv",
        syntax_case_sensitive=False,
        item_separator='__',
        default_entry_prefix='wrk',
        space_char='|',
        na_string_values=['!', 'x'],
        na_node_type='missing',
        skiprows=2,
        comment_char='#'
    )

    notna_subset = [
        'src_node_id',
        'tgt_node_id',
        'ref_node_id',
        'link_type_id',
        'date_inserted'
    ]

    assert not tn.edges[notna_subset].isna().any(axis=None)
    assert tn.edges['date_modified'].isna().all()


def test_manual_annotation_strings_column_nan_states():

    tn = bg.slurp_shorthand(
        'bibliograph/test_data/manual_annotation.shnd',
        "bibliograph/resources/default_entry_syntax.csv",
        link_syntax_fname="bibliograph/resources/default_link_syntax.csv",
        syntax_case_sensitive=False,
        item_separator='__',
        default_entry_prefix='wrk',
        space_char='|',
        na_string_values=['!', 'x'],
        na_node_type='missing',
        skiprows=2,
        comment_char='#'
    )

    notna_subset = ['node_id', 'string', 'date_inserted']

    assert not tn.strings[notna_subset].isna().any(axis=None)
    assert tn.strings['date_modified'].isna().all()


def test_manual_annotation_assertions_column_nan_states():

    tn = bg.slurp_shorthand(
        'bibliograph/test_data/manual_annotation.shnd',
        "bibliograph/resources/default_entry_syntax.csv",
        link_syntax_fname="bibliograph/resources/default_link_syntax.csv",
        syntax_case_sensitive=False,
        item_separator='__',
        default_entry_prefix='wrk',
        space_char='|',
        na_string_values=['!', 'x'],
        na_node_type='missing',
        skiprows=2,
        comment_char='#'
    )

    notna_subset = [
        'inp_string_id',
        'src_string_id',
        'tgt_string_id',
        'ref_string_id',
        'link_type_id',
        'date_inserted'
    ]

    assert not tn.assertions[notna_subset].isna().any(axis=None)
    assert tn.assertions['date_modified'].isna().all()


def test_manual_annotation_num_strings_is_65():

    tn = bg.slurp_shorthand(
        'bibliograph/test_data/manual_annotation.shnd',
        "bibliograph/resources/default_entry_syntax.csv",
        link_syntax_fname="bibliograph/resources/default_link_syntax.csv",
        syntax_case_sensitive=False,
        item_separator='__',
        default_entry_prefix='wrk',
        space_char='|',
        na_string_values=['!', 'x'],
        na_node_type='missing',
        skiprows=2,
        comment_char='#'
    )

    assert len(tn.strings) == 64


def test_manual_annotation_node_types_with_single_strings():

    tn = bg.slurp_shorthand(
        'bibliograph/test_data/manual_annotation.shnd',
        "bibliograph/resources/default_entry_syntax.csv",
        link_syntax_fname="bibliograph/resources/default_link_syntax.csv",
        syntax_case_sensitive=False,
        item_separator='__',
        default_entry_prefix='wrk',
        space_char='|',
        na_string_values=['!', 'x'],
        na_node_type='missing',
        skiprows=2,
        comment_char='#'
    )

    single_string_node_types = [
        'acknowledgement',
        'affiliation',
        'shorthand_text',
        'shorthand_entry_syntax',
        'shorthand_link_syntax'
    ]

    counts = tn.resolve_strings()['node_type'].value_counts()
    assert counts.loc[counts == 1].index.isin(single_string_node_types).all()


def test_manual_annotation_node_type_0_is_shorthand_text():

    tn = bg.slurp_shorthand(
        'bibliograph/test_data/manual_annotation.shnd',
        "bibliograph/resources/default_entry_syntax.csv",
        link_syntax_fname="bibliograph/resources/default_link_syntax.csv",
        syntax_case_sensitive=False,
        item_separator='__',
        default_entry_prefix='wrk',
        space_char='|',
        na_string_values=['!', 'x'],
        na_node_type='missing',
        skiprows=2,
        comment_char='#'
    )

    shorthand_text_node_type_id = tn.id_lookup('node_types', 'shorthand_text')

    assert (shorthand_text_node_type_id == 0)


def test_manual_annotation_single_string_has_node_type_shorthand_text():

    tn = bg.slurp_shorthand(
        'bibliograph/test_data/manual_annotation.shnd',
        "bibliograph/resources/default_entry_syntax.csv",
        link_syntax_fname="bibliograph/resources/default_link_syntax.csv",
        syntax_case_sensitive=False,
        item_separator='__',
        default_entry_prefix='wrk',
        space_char='|',
        na_string_values=['!', 'x'],
        na_node_type='missing',
        skiprows=2,
        comment_char='#'
    )

    shorthand_text_string = tn.resolve_strings().query(
        'node_type == "shorthand_text"'
    )

    assert (shorthand_text_string.node_type.squeeze() == "shorthand_text")


def test_manual_annotation_shorthand_text_string_startswith():

    tn = bg.slurp_shorthand(
        'bibliograph/test_data/manual_annotation.shnd',
        "bibliograph/resources/default_entry_syntax.csv",
        link_syntax_fname="bibliograph/resources/default_link_syntax.csv",
        syntax_case_sensitive=False,
        item_separator='__',
        default_entry_prefix='wrk',
        space_char='|',
        na_string_values=['!', 'x'],
        na_node_type='missing',
        skiprows=2,
        comment_char='#'
    )

    shorthand_text_string = tn.resolve_strings().query(
        'node_type == "shorthand_text"'
    )
    assert (shorthand_text_string.string.squeeze().startswith(
        'This is stuff shorthand ignores'
    ))


def test_aliasing_num_case_sensitive_actor_nodes_is_5():

    aliases_dict = {'actor': 'bibliograph/test_data/aliases_actor.csv'}

    tn = bg.slurp_shorthand(
        'bibliograph/test_data/shorthand_with_aliases.shnd',
        "bibliograph/resources/default_entry_syntax.csv",
        "bibliograph/resources/default_link_syntax.csv",
        syntax_case_sensitive=False,
        aliases_dict=aliases_dict,
        item_separator='__',
        space_char='|',
        na_string_values='!',
        na_node_type='missing',
        default_entry_prefix='wrk',
        skiprows=2,
        comment_char='#',
    )

    assert len(tn.resolve_nodes().query('node_type == "actor"')) == 5


def test_aliasing_num_case_sensitive_work_nodes_is_10():

    aliases_dict = {'work': 'bibliograph/test_data/aliases_work.csv'}

    tn = bg.slurp_shorthand(
        'bibliograph/test_data/shorthand_with_aliases.shnd',
        "bibliograph/resources/default_entry_syntax.csv",
        "bibliograph/resources/default_link_syntax.csv",
        syntax_case_sensitive=False,
        aliases_dict=aliases_dict,
        item_separator='__',
        space_char='|',
        na_string_values='!',
        na_node_type='missing',
        default_entry_prefix='wrk',
        skiprows=2,
        comment_char='#',
    )

    assert len(tn.resolve_nodes().query('node_type == "work"')) == 10


def test_aliasing_num_case_insensitive_actor_nodes_is_4():

    aliases_dict = {'actor': 'bibliograph/test_data/aliases_actor.csv'}

    tn = bg.slurp_shorthand(
        'bibliograph/test_data/shorthand_with_aliases.shnd',
        "bibliograph/resources/default_entry_syntax.csv",
        "bibliograph/resources/default_link_syntax.csv",
        syntax_case_sensitive=False,
        aliases_dict=aliases_dict,
        aliases_case_sensitive=False,
        item_separator='__',
        space_char='|',
        na_string_values='!',
        na_node_type='missing',
        default_entry_prefix='wrk',
        skiprows=2,
        comment_char='#',
    )

    assert len(tn.resolve_nodes().query('node_type == "actor"')) == 4


def test_aliasing_num_case_insensitive_work_nodes_is_10():

    aliases_dict = {'work': 'bibliograph/test_data/aliases_work.csv'}

    tn = bg.slurp_shorthand(
        'bibliograph/test_data/shorthand_with_aliases.shnd',
        "bibliograph/resources/default_entry_syntax.csv",
        "bibliograph/resources/default_link_syntax.csv",
        syntax_case_sensitive=False,
        aliases_dict=aliases_dict,
        aliases_case_sensitive=False,
        item_separator='__',
        space_char='|',
        na_string_values='!',
        na_node_type='missing',
        default_entry_prefix='wrk',
        skiprows=2,
        comment_char='#',
    )

    assert len(tn.resolve_nodes().query('node_type == "work"')) == 10


def test_aliasing_both_sets_num_case_insensitive_alias_assrtns_is_17():

    aliases_dict = {
        'actor': 'bibliograph/test_data/aliases_actor.csv',
        'work': 'bibliograph/test_data/aliases_work.csv'
    }

    tn = bg.slurp_shorthand(
        'bibliograph/test_data/shorthand_with_aliases.shnd',
        "bibliograph/resources/default_entry_syntax.csv",
        "bibliograph/resources/default_link_syntax.csv",
        syntax_case_sensitive=False,
        aliases_dict=aliases_dict,
        aliases_case_sensitive=False,
        item_separator='__',
        space_char='|',
        na_string_values='!',
        na_node_type='missing',
        default_entry_prefix='wrk',
        skiprows=2,
        comment_char='#',
    )

    assert len(tn.resolve_assertions().query('link_type == "alias"')) == 17


def test_aliasing_node_3_work_bams_values():

    aliases_dict = {
        'actor': 'bibliograph/test_data/aliases_actor.csv',
        'work': 'bibliograph/test_data/aliases_work.csv'
    }

    tn = bg.slurp_shorthand(
        'bibliograph/test_data/shorthand_with_aliases.shnd',
        "bibliograph/resources/default_entry_syntax.csv",
        "bibliograph/resources/default_link_syntax.csv",
        syntax_case_sensitive=False,
        aliases_dict=aliases_dict,
        aliases_case_sensitive=False,
        item_separator='__',
        space_char='|',
        na_string_values='!',
        na_node_type='missing',
        default_entry_prefix='wrk',
        skiprows=2,
        comment_char='#',
    )

    node_3_aliases = [
        'bams',
        'Bulletin of the American Meteorological Society',
        'bulletin of the american meteorological society'
    ]

    strings_with_node_3 = tn.strings.loc[tn.strings['node_id'] == 3, 'string']

    assert (strings_with_node_3 == node_3_aliases).all().all()


def test_aliasing_node_1_actor_asmith_values():
    aliases_dict = {
        'actor': 'bibliograph/test_data/aliases_actor.csv',
        'work': 'bibliograph/test_data/aliases_work.csv'
    }

    tn = bg.slurp_shorthand(
        'bibliograph/test_data/shorthand_with_aliases.shnd',
        "bibliograph/resources/default_entry_syntax.csv",
        "bibliograph/resources/default_link_syntax.csv",
        syntax_case_sensitive=False,
        aliases_dict=aliases_dict,
        aliases_case_sensitive=False,
        item_separator='__',
        space_char='|',
        na_string_values='!',
        na_node_type='missing',
        default_entry_prefix='wrk',
        skiprows=2,
        comment_char='#',
    )

    node_1_aliases = ['asmith', 'Alice Smith', 'alice smith']

    strings_with_node_1 = tn.strings.loc[tn.strings['node_id'] == 1, 'string']

    assert (strings_with_node_1 == node_1_aliases).all().all()


def test_aliasing_node_2_actor_bwu_values():
    aliases_dict = {
        'actor': 'bibliograph/test_data/aliases_actor.csv',
        'work': 'bibliograph/test_data/aliases_work.csv'
    }

    tn = bg.slurp_shorthand(
        'bibliograph/test_data/shorthand_with_aliases.shnd',
        "bibliograph/resources/default_entry_syntax.csv",
        "bibliograph/resources/default_link_syntax.csv",
        syntax_case_sensitive=False,
        aliases_dict=aliases_dict,
        aliases_case_sensitive=False,
        item_separator='__',
        space_char='|',
        na_string_values='!',
        na_node_type='missing',
        default_entry_prefix='wrk',
        skiprows=2,
        comment_char='#',
    )

    node_2_aliases = [
        'bwu', 'Elizabeth Wu', 'Beth Wu', 'beth wu', 'elizabeth wu'
    ]

    strings_with_node_2 = tn.strings.loc[tn.strings['node_id'] == 2, 'string']

    assert (strings_with_node_2 == node_2_aliases).all().all()


def test_aliasing_node_0_actor_nasa_values():
    aliases_dict = {
        'actor': 'bibliograph/test_data/aliases_actor.csv',
        'work': 'bibliograph/test_data/aliases_work.csv'
    }

    tn = bg.slurp_shorthand(
        'bibliograph/test_data/shorthand_with_aliases.shnd',
        "bibliograph/resources/default_entry_syntax.csv",
        "bibliograph/resources/default_link_syntax.csv",
        syntax_case_sensitive=False,
        aliases_dict=aliases_dict,
        aliases_case_sensitive=False,
        item_separator='__',
        space_char='|',
        na_string_values='!',
        na_node_type='missing',
        default_entry_prefix='wrk',
        skiprows=2,
        comment_char='#',
    )

    node_0_aliases = [
        'NASA',
        'National Aeronautics and Space Administration',
        'nasa',
        'national aeronautics and space administration'
    ]

    strings_with_node_0 = tn.strings.loc[tn.strings['node_id'] == 0, 'string']

    assert (strings_with_node_0 == node_0_aliases).all().all()


def test_auto_aliasing_lemonem_and_personso_assertions():

    aliases_dict = {
        'actor': 'bibliograph/test_data/aliases_actor.csv',
        'work': 'bibliograph/test_data/aliases_work.csv'
    }

    constraints_fname = "bibliograph/resources/default_link_constraints.csv"

    tn = bg.slurp_shorthand(
        'bibliograph/test_data/shorthand_for_auto_aliasing.shnd',
        "bibliograph/resources/default_entry_syntax.csv",
        "bibliograph/resources/default_link_syntax.csv",
        syntax_case_sensitive=False,
        aliases_dict=aliases_dict,
        aliases_case_sensitive=False,
        automatic_aliasing=True,
        link_constraints_fname=constraints_fname,
        item_separator='__',
        space_char='|',
        na_string_values='!',
        na_node_type='missing',
        default_entry_prefix='wrk',
        comment_char='#',
    )

    input_string = (
        "bibliograph.core.slurp_shorthand(**{'shorthand_fname': "
        "'bibliograph/test_data/shorthand_for_auto_aliasing.shnd', "
        "'entry_syntax_fname': "
        "'bibliograph/resources/default_entry_syntax.csv', "
        "'link_syntax_fname': 'bibliograph/resources/default_link_syntax.csv',"
        " 'syntax_case_sensitive': False, 'allow_redundant_items': False, "
        "'aliases_dict': {'actor': 'bibliograph/test_data/aliases_actor.csv', "
        "'work': 'bibliograph/test_data/aliases_work.csv'}, "
        "'aliases_case_sensitive': False, 'automatic_aliasing': True, "
        "'link_constraints_fname': "
        "'bibliograph/resources/default_link_constraints.csv', "
        "'links_excluded_from_edges': None, 'item_separator': '__', "
        "'space_char': '|', 'na_string_values': '!', 'na_node_type': "
        "'missing', 'default_entry_prefix': 'wrk', 'comment_char': '#'})"
    )

    correct_assertions = pd.DataFrame(
        {
            'inp_string': input_string,
            'src_string': ['margaret lemone', 's. o. person'],
            'tgt_string': ['lemonem', 'personso'],
            'link_type': 'alias',
            'inp_node_type': 'python_function',
            'src_node_type': 'actor',
            'tgt_node_type': 'actor',
            'ref_node_type': 'alias_reference'
        },
        index=pd.Index([76, 79]).astype(tn.big_id_dtype)
    )

    assertion_selection = tn.resolve_assertions().loc[
        correct_assertions.index,
        correct_assertions.columns
    ]

    assert (assertion_selection == correct_assertions).all().all()


def test_auto_aliasing_by_stitle_vol_pg_link_constraints():

    aliases_dict = {
        'actor': 'bibliograph/test_data/aliases_actor.csv',
        'work': 'bibliograph/test_data/aliases_work.csv'
    }

    constraints_fname = "bibliograph/resources/default_link_constraints.csv"

    tn = bg.slurp_shorthand(
        'bibliograph/test_data/shorthand_for_auto_aliasing.shnd',
        "bibliograph/resources/default_entry_syntax.csv",
        "bibliograph/resources/default_link_syntax.csv",
        syntax_case_sensitive=False,
        aliases_dict=aliases_dict,
        aliases_case_sensitive=False,
        automatic_aliasing=True,
        link_constraints_fname=constraints_fname,
        item_separator='__',
        space_char='|',
        na_string_values='!',
        na_node_type='missing',
        default_entry_prefix='wrk',
        comment_char='#',
    )

    node_10_aliases = [
        'asmith_bwu__1999__bams__101__803__xxx',
        'smitha_wub__1999__bams__101__803'
    ]
    strings_with_node_10 = tn.strings.loc[
        tn.strings['node_id'] == 10,
        'string'
    ]

    assert (strings_with_node_10 == node_10_aliases).all().all()


def test_auto_aliasing_by_doi_link_constraints():

    aliases_dict = {
        'actor': 'bibliograph/test_data/aliases_actor.csv',
        'work': 'bibliograph/test_data/aliases_work.csv'
    }

    constraints_fname = "bibliograph/resources/default_link_constraints.csv"

    tn = bg.slurp_shorthand(
        'bibliograph/test_data/shorthand_for_auto_aliasing.shnd',
        "bibliograph/resources/default_entry_syntax.csv",
        "bibliograph/resources/default_link_syntax.csv",
        syntax_case_sensitive=False,
        aliases_dict=aliases_dict,
        aliases_case_sensitive=False,
        automatic_aliasing=True,
        link_constraints_fname=constraints_fname,
        item_separator='__',
        space_char='|',
        na_string_values='!',
        na_node_type='missing',
        default_entry_prefix='wrk',
        comment_char='#',
    )

    node_14_aliases = [
        'Alice Smith_Elizabeth Wu__1998__bams__100__42__yyy',
        'smitha_wub__1998__!__!__!__doi:yyy'
    ]
    strings_with_node_14 = tn.strings.loc[
        tn.strings['node_id'] == 14,
        'string'
    ]

    assert (strings_with_node_14 == node_14_aliases).all().all()


def test_s_d_strings_subset_same_as_shnd_strings_subset():

    s = bg.Shorthand(
        entry_syntax="bibliograph/resources/default_entry_syntax.csv",
        link_syntax="bibliograph/resources/default_link_syntax.csv",
        syntax_case_sensitive=False
    )

    shorthand_text = (
        'left_entry, right_entry, link_tags_or_override, reference\n'
        'Auth1__2000__BAMS__100__!__xxx__ tag1 tag2, '
        'Auth2__1990__BAMS__90__40__yyy__ tag1 tag3, '
        'LinkTag'
    )

    self_descriptive = (
        'left_entry, right_entry, link_tags_or_override, reference\n'
        '____work__author_actor_Auth1__published_date_2000__supertitle_work_'
        'BAMS__volume_work_100__page_work_!__doi_identifier_xxx__ tag1 tag2, '
        '____work__author_actor_Auth2__published_date_1990__supertitle_work_'
        'BAMS__volume_work_90__page_work_40__doi_identifier_yyy__ tag1 tag3, '
        'lt__cited LinkTag'
    )

    parsed_shnd_sources = s.parse_text(
        StringIO(shorthand_text),
        item_separator='__',
        default_entry_prefix='wrk',
        space_char='|',
        na_string_values='!',
        na_node_type='missing',
        comment_char='#'
    )

    parsed_s_d_sources = s.parse_text(
        StringIO(self_descriptive),
        item_separator='__',
        default_entry_prefix='wrk',
        space_char='|',
        na_string_values='!',
        na_node_type='missing',
        comment_char='#'
    )

    node_types = [
        'entry',
        'shorthand_text',
        'shorthand_link_syntax',
        'shorthand_entry_syntax',
        'work'
    ]

    shnd_strings = parsed_shnd_sources.resolve_strings()
    shnd_strings = shnd_strings.loc[
        ~shnd_strings['node_type'].isin(node_types)
    ]
    shnd_strings = shnd_strings.loc[shnd_strings['string'] != '1']
    shnd_strings = shnd_strings.sort_values(by='string').reset_index(drop=True)

    s_d_strings = parsed_s_d_sources.resolve_strings()
    s_d_strings = s_d_strings.loc[~s_d_strings['node_type'].isin(node_types)]
    s_d_strings = s_d_strings.sort_values(by='string').reset_index(drop=True)

    assert (shnd_strings == s_d_strings).all().all()


def test_bibtex_identifier_parsing():

    bibtex_parser = BibTexParser(common_strings=True)
    bibtex_fname = "bibliograph/test_data/bibtex_test_data_short.bib"
    with open(bibtex_fname, encoding='utf8') as f:
        bibdatabase = bibtex_parser.parse_file(f)

    data = pd.DataFrame(bibdatabase.entries)

    s = bg.Shorthand(
        entry_syntax="bibliograph/resources/default_bibtex_syntax.csv",
        syntax_case_sensitive=False
    )

    parsed = s.parse_items(
        data,
        space_char='|',
        na_string_values='!',
        na_node_type='missing'
    )

    parsed_identifiers = parsed.strings.query('node_type_id == 8')

    check = pd.Series([
        '10.1038/194638b0',
        '10.1175/1520-0493(1962)090<0311:OTOKEB>2.0.CO;2',
        '10.3402/tellusa.v14i3.9551',
        '10.1175/1520-0477-43.9.451',
        '10.3402/tellusa.v14i4.9569',
        '10.1007/BF02317953',
        '10.1007/BF02247180',
        '10.1029/JZ068i011p03345',
        '10.1029/JZ068i009p02375',
    ])

    assert (check == parsed_identifiers['string'].array).all()


def test_parsed_bibtex_items_resolve_links_has_no_nans():

    bibtex_parser = BibTexParser(common_strings=True)
    bibtex_test_data_fname = "bibliograph/test_data/bibtex_test_data_short.bib"
    with open(bibtex_test_data_fname, encoding='utf8') as f:
        bibdatabase = bibtex_parser.parse_file(f)

    data = pd.DataFrame(bibdatabase.entries)

    s = bg.Shorthand(
        entry_syntax="bibliograph/resources/default_bibtex_syntax.csv",
        syntax_case_sensitive=False
    )

    parsed = s.parse_items(
        data.iloc[:4],
        space_char='|',
        na_string_values='!',
        na_node_type='missing'
    )

    assert (~parsed.resolve_links().isna()).any().any()


def test_single_column_wrk_synthesis():

    s = bg.Shorthand(
        entry_syntax="bibliograph/resources/default_entry_syntax.csv",
        link_syntax="bibliograph/resources/default_link_syntax.csv",
        syntax_case_sensitive=False
    )

    parsed = s.parse_text(
        'bibliograph/test_data/single_column.shnd',
        item_separator='__',
        default_entry_prefix='wrk',
        space_char='|',
        na_string_values=['!', 'x'],
        na_node_type='missing',
        skiprows=0,
        comment_char='#',
        drop_na=False
    )

    synthesized = parsed.synthesize_shorthand_entries('wrk', fill_spaces=True)

    check = pd.Series([
        'asmith_bwu__1999__s_bams__101__803__xxx',
        'asmith_bwu__1998__s_bams__100__42__yyy',
        'bjones__1975__s_jats__90__1__!',
        'bwu__1989__t_long|title__x__80__!',
        'Some|Author__1989__t_A|Title|With|\\#__x__x__!',
        'asmith_bwu__2008__s_bams__110__1__zzz'
    ])

    assert (check == synthesized).all()


def test_manual_annotation_note_synthesis():

    s = bg.Shorthand(
        entry_syntax="bibliograph/resources/default_entry_syntax.csv",
        link_syntax="bibliograph/resources/default_link_syntax.csv",
        syntax_case_sensitive=False
    )

    parsed = s.parse_text(
        'bibliograph/test_data/manual_annotation.shnd',
        item_separator='__',
        default_entry_prefix='wrk',
        space_char='|',
        na_string_values=['!', 'x'],
        na_node_type='missing',
        skiprows=2,
        comment_char='#'
    )

    synthesized = parsed.synthesize_shorthand_entries(entry_node_type='note')

    check = pd.Series([
        'this is an article I made up for testing',
        'here\'s a note with an escaped\\__item separator and some '
        '"quotation marks"'
    ])

    assert (check == synthesized).all()


def test_manual_annotation_wrk_synthesis():

    s = bg.Shorthand(
        entry_syntax="bibliograph/resources/default_entry_syntax.csv",
        link_syntax="bibliograph/resources/default_link_syntax.csv",
        syntax_case_sensitive=False
    )

    parsed = s.parse_text(
        'bibliograph/test_data/manual_annotation.shnd',
        item_separator='__',
        default_entry_prefix='wrk',
        space_char='|',
        na_string_values=['!', 'x'],
        na_node_type='missing',
        skiprows=2,
        comment_char='#'
    )

    synthesized = parsed.synthesize_shorthand_entries('wrk', fill_spaces=True)

    check = pd.Series([
        'asmith_bwu__1999__s_bams__101__803__xxx',
        'asmith_bwu__1998__s_bams__100__42__yyy',
        'bjones__1975__s_jats__90__1__!',
        'bwu__1989__t_long|title__x__80__!',
        'Some|Author__1989__t_Title|With|\\#__x__x__!',
        'asmith_bwu__2008__s_bams__110__1__zzz'
    ])

    assert (check == synthesized).all()


def test_parsed_manual_annotation_resolve_links_has_no_nans():

    s = bg.Shorthand(
        entry_syntax="bibliograph/resources/default_entry_syntax.csv",
        link_syntax="bibliograph/resources/default_link_syntax.csv",
        syntax_case_sensitive=False
    )

    parsed = s.parse_text(
        'bibliograph/test_data/manual_annotation.shnd',
        item_separator='__',
        default_entry_prefix='wrk',
        space_char='|',
        na_string_values=['!', 'x'],
        na_node_type='missing',
        skiprows=2,
        comment_char='#'
    )

    assert (~parsed.resolve_links().isna()).any().any()



import pandas as pd


def get_string_values(
    obj,
    string_subset=slice(None),
    node_type_subset=slice(None),
    casefold=False
):
    if obj.strings.empty:
        return obj.strings['string']

    strings_idx = obj.strings['string'].loc[string_subset].index

    try:
        assert node_type_subset.casefold()

        node_type_id = obj.id_lookup('node_types', node_type_subset)
        node_types_idx = obj.strings.query(
            'node_type_id == @node_type_id'
        )
        node_types_idx = node_types_idx.index

    except AttributeError:

        try:
            node_type_id = int(node_type_subset)
            node_types_idx = obj.strings.query(
                'node_type_id == @node_type_id'
            )
            node_types_idx = node_types_idx.index

        except ValueError:
            node_types_idx = obj.strings['node_type_id'].loc[
                node_type_subset
            ]
            node_types_idx = node_types_idx.index

    selection = strings_idx.intersection(node_types_idx)
    selection = obj.strings.loc[selection, 'string']

    if casefold:
        return selection.str.casefold()

    else:
        return selection


def make_bidirectional_map_one_to_many(df):

    input_columns = list(df.columns)
    df.columns = [0, 1]

    # Drop identities in a way that doesn't copy the whole dataframe.
    identities = df.loc[df[0] == df[1]]
    df = df.drop(identities.index).drop_duplicates()
    # Using the above method because the following line raises a copy
    # vs. view warning:
    # df = df.loc[df[0] != df[1], :]

    # Check if the input dataframe is already a one-to-many map
    if not df[1].duplicated().any() and not df[1].isin(df[0]).any():
        df.columns = input_columns
        return df

    # Stack the values so we can find duplicate values regardless of the
    # column
    values = df.stack()
    duplicate_values = values.loc[values.duplicated(keep=False)]

    '''
    Drop rows of the form [b, a] when the row [a, b] is also present
    '''

    # Get values that are duplicated and present in both columns
    value_selection = duplicate_values.loc[
        duplicate_values.isin(df[0]) & duplicate_values.isin(df[1])
    ]
    drop_candidates = df.loc[df[1].isin(value_selection)]

    # Hash the dataframe to search for whole rows
    hashed_df = pd.util.hash_pandas_object(df, index=False)

    # Reverse the selected rows and hash them to search for the
    # reversed rows in the input dataframe
    reversed_candidates = drop_candidates[[1, 0]]
    hashed_candidates = pd.util.hash_pandas_object(
        reversed_candidates,
        index=False
    )

    # Drop the reversed rows if they're present in the dataframe
    drop_candidates = hashed_df.isin(hashed_candidates)
    if drop_candidates.any():

        # If the search above caught [a, b] then it also caught [b, a]
        # so we select one of those rows by stacking, sorting by index
        # and then by value, pivoting back to the original shape, and
        # then dropping duplicate rows.
        drop_candidates = df.loc[hashed_df[drop_candidates].index]
        drop_candidates = drop_candidates.stack().reset_index()
        drop_candidates = drop_candidates.sort_values(by=['level_0', 0])
        drop_candidates.loc[:, 'level_1'] = [0, 1]*int(len(drop_candidates)/2)
        drop_candidates = drop_candidates.pivot(
            index='level_0',
            columns='level_1',
            values=0
        )

        drop_rows = drop_candidates.drop_duplicates().index

        df = df.drop(drop_rows)

        # If the dataframe is now a one-to-many map, we're done
        if not df[1].duplicated().any():
            if not df[1].isin(df[0]).any():
                df.columns = input_columns
                return df

        # Stack the modified dataframe and get duplicate values
        values = df.stack()
        duplicate_values = values.loc[values.duplicated(keep=False)]

    '''
    Locate two different types of duplicate values in the second column
    '''

    # Get first instance in the second column of values duplicated
    # anywhere in the dataframe
    first_instance_of_any_dplct_in_c1 = (
        duplicate_values.loc[(slice(None), 1)].drop_duplicates()
    )

    # Get values from the second column which are duplicated anywhere
    # in the dataframe but which occurred first in the second column
    frst_instnc_of_dplct_found_in_c1_frst = duplicate_values.drop_duplicates()
    if 1 in frst_instnc_of_dplct_found_in_c1_frst.index.get_level_values(1):
        frst_instnc_of_dplct_found_in_c1_frst = (
            frst_instnc_of_dplct_found_in_c1_frst.loc[(slice(None), 1)]
        )

    else:
        frst_instnc_of_dplct_found_in_c1_frst = pd.Series([], dtype=int)

    # If values duplicated anywhere in the dataframe exist in the
    # second column, map the duplicates to the first-column
    # partner of the first instance of the value in the second column
    if df[1].duplicated().any():

        '''
        Make maps between the different kinds of first second-column
        instances and their first-column partners
        '''

        # Map from the first instance in the second column of values
        # duplicated anywhere in the dataframe to their first-column
        # partners
        df_selection = df.loc[first_instance_of_any_dplct_in_c1.index]
        map_to_c0_from_frst_instnc_of_any_dplct_in_c1 = pd.Series(
            df_selection[0].array,
            index=df_selection[1].array
        )

        # Map from duplicate values that occurred first in the second
        # column to the first-column partners of their first instance
        map_to_c0_from_frst_instnc_of_dplct_found_in_c1_frst = df.loc[
            frst_instnc_of_dplct_found_in_c1_frst.index,
            :
        ]
        map_to_c0_from_frst_instnc_of_dplct_found_in_c1_frst = pd.Series(
            map_to_c0_from_frst_instnc_of_dplct_found_in_c1_frst[0].array,
            index=map_to_c0_from_frst_instnc_of_dplct_found_in_c1_frst[1].array
        )

        '''
        Apply the maps
        '''

        # Get rows with values in the first column that appeared
        # first in the second column
        c0_value_was_found_in_c1_first = df.loc[
            df[0].isin(first_instance_of_any_dplct_in_c1),
            :
        ]

        # Map those values to the first-column partners of their first
        # instances in the second column
        mapped_c0 = c0_value_was_found_in_c1_first[0].map(
            map_to_c0_from_frst_instnc_of_any_dplct_in_c1
        )
        df.loc[c0_value_was_found_in_c1_first.index, 0] = mapped_c0.array

        # Get all duplicate values in the second column which appeared first
        # in the second column
        c1_value_was_found_in_c1_first = df.loc[
            df[1].isin(frst_instnc_of_dplct_found_in_c1_frst),
            :
        ]

        # Drop the first instances to isolate only the duplicates
        c1_value_was_found_in_c1_first = c1_value_was_found_in_c1_first.drop(
            frst_instnc_of_dplct_found_in_c1_frst.index
        )

        # If there are duplicate values in the second column which
        # appeared first in the second column, map those values to the
        # first-column partners of their first instances and then swap
        # values in those rows so the first instance's first-column
        # partner is in the first column

        if not c1_value_was_found_in_c1_first.empty:
            mapped_1 = c1_value_was_found_in_c1_first[1].map(
                map_to_c0_from_frst_instnc_of_dplct_found_in_c1_frst
            )
            df.loc[c1_value_was_found_in_c1_first.index, :] = list(zip(
                mapped_1.array,
                df.loc[c1_value_was_found_in_c1_first.index, 0]
            ))
            df = df.drop_duplicates()

    # Rearrange the dataframe so that values which appear in both
    # columns appear first in the second column

    c1_in_c0 = df[1].isin(df[0])
    df = pd.concat([df.loc[c1_in_c0], df.loc[~c1_in_c0]])

    c0_in_c1 = df[0].isin(df[1])
    df = pd.concat([df.loc[~c0_in_c1], df.loc[c0_in_c1]])

    # In some cases the above algorithm will loop indefinitely if there
    # are values that occur in both columns but there are no repeated
    # values in the second column. Avoid this finding rows with values
    # in the first column that are present in the second column and, if
    # any exist, swapping values in the first row found.
    c0_in_c1 = df[0].isin(df[1])

    if c0_in_c1.any():
        to_swap = df.loc[c0_in_c1].index[0]
        df.loc[to_swap, :] = df.loc[to_swap, [1, 0]].to_numpy()

    df.columns = input_columns

    # Remap the modified dataframe
    return make_bidirectional_map_one_to_many(df)


def escape_regex_metachars(s):
    s = s.replace("\\", "\\\\")
    metachars = '.^$*+?{}[]|()'
    for metachar in [c for c in metachars if c in s]:
        s = s.replace(metachar, f'\\{metachar}')
    return s


def get_single_value(
    df_or_s,
    label,
    none_ok=False,
    group_key=None
):

    try:
        # If this assertion passes then assume input is a pandas.Series
        assert df_or_s.str

        if df_or_s.empty:
            raise ValueError(f'Series with label "{label}" is empty.')
        elif len(df_or_s) == 1:
            return df_or_s.array[0]

        values = df_or_s.value_counts()

        num_values = len(values)

        if num_values == 1:
            return values.index[0]
        elif num_values > 1:
            raise ValueError(f'Found multiple values with label "{label}"')
        elif num_values < 1:
            if not none_ok:
                raise ValueError(f'Found only nan values with label "{label}"')
            else:
                return None

    except AttributeError:

        # If the assertion failed then assume input is a pandas.DataFrame
        if len(df_or_s[label]) == 1:
            return df_or_s[label].array[0]

        if group_key is None:
            message_tail = f'in column "{label}"'

        else:
            message_tail = 'in column "{}", group key "{}"'.format(
                label,
                group_key
            )

        values = df_or_s[label].value_counts()

        num_values = len(values)

        if num_values == 1:
            return values.index[0]
        elif num_values > 1:
            raise ValueError('Found multiple values {}'.format(message_tail))
        elif num_values < 1:
            if not none_ok:
                raise ValueError('No values found {}'.format(message_tail))
            else:
                return None


def iterable_not_string(x):
    '''
    Check if input has an __iter__ method and then determine if it's a
    string by checking for a casefold method.
    '''
    try:
        assert x.__iter__

        try:
            assert x.casefold
            return False

        except AttributeError:
            return True

    except AttributeError:
        return False


def normalize_types(to_norm, template, strict=True, continue_idx=True):
    '''
    Create an object from to_norm that can be concatenated with template
    such that the concatenated object and its index will have the same
    dtypes as the template.

    If template is pandas.DataFrame and to_norm is dict-like or
    list-like, convert each element of to_norm to a pandas.Series with
    dtypes that conform to the dtypes of a template dataframe and and
    concatenate them together as columns in a dataframe.

    If template is pandas.Series, convert to_norm to a Series with the
    appropriate array dtype or, if to_norm is dict-like,
    convert its values to an array with the appropriate dtype and if
    both to_norm and template have numeric indexes, also normalize the
    index dtype. If to_norm is dict-like and either to_norm or template
    does not have a numeric index, use to_norm.keys() as the output
    index.

    Parameters
    ----------
    to_norm : dict-like or list-like
        A set of objects to treat as columns of an output
        pandas.DataFrame and whose types will be coerced.

    template : pandas.DataFrame or pandas.Series
        Output dtypes will conform to template.dtypes and the output
        index dtype will be template.index.dtype

    strict : bool, default True
        If True and to_norm is dict-like, only objects whose keys are
        column labels in the template will be included as columns in the
        output dataframe.

        If True, to_norm is list-like, and template has N columns,
        include only the first N elements of to_norm in the output
        dataframe.

        If False and to_norm is dict-like, normalize dtypes for objects
        whose keys are column labels in the template and include the
        other elements of to_norm as columns with dtypes inferred by
        the pandas.Series constructor.

        If False, to_norm is list-like, and template has N columns,
        normalize dtypes for the first N elements of to_norm and include
        the other elements of to_norm as columns with dtypes inferred by
        the pandas.Series constructor. Labels for the extra columns in
        the output dataframe will be integers counting from N.

    continue_index : bool, default True
        If True and template has a numerical index, the index of the
        returned object will be a sequence of integers which fills
        gaps in and/or extends to_norm.index

    Returns
    -------
    pandas.DataFrame
        Has as many rows as the longest element of to_norm.

        If to_norm is dict-like and strict is True, output includes
        only objects in to_norm whose keys are also column labels in the
        template.

        If to_norm is list-like and strict is True, output has the same
        width as the template.

        If strict is False, output has one column for each element in
        to_norm.
    '''

    # get the size of to_norm

    try:
        # to_norm is treated as columns, so if it has a columns
        # attribute, we want the size of that instead of the length
        num_elements = len(to_norm.columns)
    except AttributeError:
        num_elements = len(to_norm)

    # check if the template index has a numeric dtype
    templt_idx_is_nmrc = pd.api.types.is_numeric_dtype(template.index)

    try:
        tmplt_columns = template.columns
        # If the template is a dataframe, get its width.
        num_tmplt_columns = len(tmplt_columns)

    except AttributeError:
        # Template is not dataframe-like.
        # Treat it as 1D object with attributes dtype and index

        try:
            assert to_norm.items
            # to_norm is dict-like, so process its keys as an index

            norm_keys_are_nmrc = pd.api.types.is_numeric_dtype(
                pd.Series(to_norm.keys())
            )
            if strict and norm_keys_are_nmrc and templt_idx_is_nmrc:
                index = non_intersecting_sequence(
                    to_norm.keys(),
                    template.index
                )
            else:
                index = to_norm.keys()

            try:
                # values is callable for a dict-like but not a
                # pandas.Series
                values = to_norm.values()

            except TypeError:
                values = to_norm.array

        except AttributeError:
            if strict and templt_idx_is_nmrc:
                index = non_intersecting_sequence(num_elements, template.index)
            else:
                index = range(num_elements)

            values = to_norm

        index = pd.Index(index, dtype=template.index.dtype)
        return pd.Series(values, index=index, dtype=template.dtype)

    try:
        # check if to_norm is dict-like
        assert to_norm.items
        # if so, optionally get extra columns
        if (num_elements > num_tmplt_columns) and not strict:
            extra_columns = {
                k: v for k, v in to_norm.items() if k not in tmplt_columns
            }

    except AttributeError:
        # to_norm is not dict-like
        if num_elements < num_tmplt_columns:
            raise ValueError(
                'If to_norm is list-like, to_norm must have at least '
                'as many elements as there are columns in template.'
            )
        # optionally get extra columns
        if (num_elements > num_tmplt_columns) and not strict:
            extra_columns = dict(zip(
                range(num_tmplt_columns, num_elements),
                to_norm[num_tmplt_columns:]
            ))
        # make the list-like dict-like
        to_norm = dict(zip(tmplt_columns, to_norm[:num_tmplt_columns]))

    to_norm = [
        pd.Series(v, dtype=template[k].dtype, name=k)
        for k, v in to_norm.items() if k in tmplt_columns
    ]

    try:
        to_norm += [pd.Series(v, name=k) for k, v in extra_columns.items()]
    except NameError:
        pass

    new_df = pd.concat(to_norm, axis='columns')

    if continue_idx and templt_idx_is_nmrc:
        template_idx_is_sequential = pd.Series(template.index).diff().iloc[1:]
        template_idx_is_sequential = (template_idx_is_sequential == 1).all()

        if template_idx_is_sequential & template.index.is_monotonic:
            new_index_min = template.index.max() + 1
            index = pd.RangeIndex(new_index_min, new_index_min + len(new_df))
        else:
            index = non_intersecting_sequence(new_df.index, template.index)

        new_df.index = pd.Index(index, dtype=template.index.dtype)

    return new_df


def get_new_typed_values(
    candidate_values,
    existing_values,
    value_column_label,
    type_column_label,
    candidate_type
):
    '''
    Selects values out of a Series of candidate values if the values are
    either not present in an existing dataframe or they are present but
    do not have the same type as the candidates.

    Parameters
    ----------
    candidate_values : pandas.Series
        Values which may or may not be present in an existing dataframe

    existing_values : pandas.DataFrame
        A dataframe with a column of values to select against and a
        column that identifies the type of each values

    value_column_label
        The label of the column of values in the existing_values
        dataframe

    type_column_label
        The label of the column of types in the existing_values
        dataframe

    candidate_type
        The type of the candidate values

    Returns
    -------
    pandas.Series
        Subset of the candidate values which are either not present in
        the set of existing values or which are present but have a
        different type in the set of existing values
    '''
    candidate_already_exists = existing_values[value_column_label].isin(
        candidate_values
    )
    wrong_type = existing_values[type_column_label] != candidate_type
    candidate_exists_wrong_type = candidate_already_exists & wrong_type

    candidate_is_new_type = candidate_values.isin(
        existing_values.loc[candidate_exists_wrong_type, value_column_label]
    )

    candidate_is_new_value = ~candidate_values.isin(
        existing_values[value_column_label]
    )

    return candidate_values.loc[candidate_is_new_type | candidate_is_new_value]


def missing_integers(input_values, rng=None):
    '''
    Creates a set of integers in a target range that does
    not intersect with values in an input set. Default range
    fills gaps from 0 to the largest input value.

    If target range does not intersect with input, return
    target.

    If target range exactly covers or is a subset of input,
    return empty set.

    input
    -----
    input_values (list-like or set):
        integers to exclude from output

    output
    ------
    (set):
        integers in target range that are not in input set
    '''
    if len(rng) != 2:
        raise ValueError('rng must be list-like with two elements')

    input_values = set(input_values)

    if rng is None:
        rng = [0, int(max(input_values)) + 1]

    target_range = set(range(*rng))

    if target_range.intersection(input_values) == set():
        return target_range
    if target_range.union(input_values) == input_values:
        return set()
    else:
        return set(input_values ^ target_range)


def non_intersecting_sequence(
    to_map,
    existing=[],
    rng=None,
    full=False,
    ignore_duplicates=False
):
    '''
    Maps unique values from a new list-like or set of integers onto the
    smallest sequence of integers that does not intersect with unique
    values in an existing list-like or set of integers.

    inputs
    ------
    to_map (list-like, set, or integer):
        integers to map to integers excluding existing integers. if an
        integer or float with integer value, to_map is converted to a
        range of that length
    existing (list-like or set):
        integers to keep
    rng (integer-valued list-like):
        range of target values for mapping
    full (bool):
        if true, return mapped values appended to existing values
        if false, return only mapped values
    ignore_duplicates (bool):
        if true, map every element in the input set to a unique value in
        the output

    output
    ------
    pandas array:
        existing and mapped integers

    examples
    --------
    a = [2,3,3]
    b = [4,3,4]
    non_intersecting_sequence(a, b)
    # output: Series([0,1,0])

    a = [2,3,3]
    b = [4,3,4]
    non_intersecting_sequence(a, b, rng=[-2,None], full=True)
    # output: Series([4,3,4,-2,-1,-2])

    c = [9,10,9]
    d = [0,2,4]
    non_intersecting_sequence(c, d, ignore_duplicates=True)
    # output: Series([1,3,5])
    '''

    if type(existing) == set:
        existing = list(existing)

    existing = pd.Series(existing, dtype='int')

    if type(to_map) == set:
        to_map = list(to_map)
    elif (type(to_map) == float) and not to_map.is_integer():
        raise ValueError('to_map cannot be a non-integer float')
    elif not iterable_not_string(to_map):
        try:
            int(to_map)
        except TypeError:
            pass
        to_map = range(to_map)

    to_map = pd.Series(to_map, dtype='int')
    unique_ints_to_map = to_map.drop_duplicates()
    len_of_req_sqnce = len(unique_ints_to_map)

    # NOTE
    # I initially had a default value rng=[0, None] in the function
    # signature. In prototyping with Jupyter lab, multiple calls to this
    # function in a row would retain the previous value of rng (the
    # second element would not be None in the second call). I don't know
    # why this happened and I don't know why using rng=None as a default
    # in the signature fixed the problem.

    if rng is None:
        if existing.empty:
            rng = [0, len_of_req_sqnce + 1]
        else:
            rng = [0, int(existing.max()) + len_of_req_sqnce + 1]

    elif len(rng) != 2:
        raise ValueError('rng must be list-like with two elements')

    elif len(range(*rng)) < (len(unique_ints_to_map) + 1):
        raise ValueError(
            'range(*rng) must include at least as many values as '
            'number of values to map'
        )

    available_ints = missing_integers(existing.drop_duplicates(), rng)
    available_ints = pd.Series(list(available_ints), dtype='int')
    available_ints = available_ints.sort_values().reset_index(drop=True)

    if ignore_duplicates:
        new_ints = available_ints[:len(to_map)]
    else:
        new_ints = available_ints.iloc[:len_of_req_sqnce]
        int_map = dict(zip(unique_ints_to_map, new_ints))
        new_ints = to_map.apply(lambda x: int_map[x])

    if full:
        full_sequence = pd.Series(
            existing.append(new_ints, ignore_index=True),
            dtype='int'
        )
        return full_sequence.array
    else:
        return pd.Series(new_ints, dtype='int').array


def set_string_dtype(df):
    '''
    can't use pd.StringDtype() throughout because it currently doesn't
    allow construction with null types other than pd.NA. This will
    likely change soon
    https://github.com/pandas-dev/pandas/pull/41412
    '''
    df_cols = df.columns
    return df.astype(
        dict(
            zip(
                df_cols, [pd.StringDtype()]*len(df_cols)
            )
        )
    )



import setuptools

with open('README.md', 'r') as readme:
    description = readme.read()

setuptools.setup(
    name='bibliograph',
    version='0.0.1',
    author='Devin Short',
    author_email='short.devin@gmail.com',
    packages=['bibliograph'],
    description=(
        'A database system for research in the humanities'
    ),
    long_description=description,
    long_description_content_type="text/markdown",
    url='https://github.com/shortorian/bibliograph',
    license='MIT',
    python_requires='>=3',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English'
    ],
    install_requires=[
        'bibtexparser>=1.3.0',
        'numpy>=1.23.1',
        'pandas>=1.4.3',
        'pyparsing>=3.0.9',
        'python-dateutil>=2.8.2',
        'pytz>=2022.1',
        'six>=1.16.0'
    ]
)
